import finesse


def test_make_micro_resonator():
    """Tests that the micro resonator example still builds and runs no physics checks
    here."""
    import numpy as np
    from finesse.components.workspace import ConnectorWorkspace
    from finesse.parameter import float_parameter
    from finesse.components import Connector, NodeType

    class MicroResonatorWorkspace(ConnectorWorkspace):
        pass

    @float_parameter("mass", "Mass", units="kg")
    @float_parameter("f", "Frequency", units="Hz")
    @float_parameter("Q", "Q-factor")
    class MicroResonator(Connector):
        """A micro resonator mechanical object.

        This mechanics element introduces three mechanical frequencies, one at the signal
        frequency and another at the resonance frequency +- the signal frequency.

        This element is meant to represent a micro resonantor whos resonance
        frequency matches the frequency separation between two carriers incident upon it.

        The object being suspended must have a mechanical port with
        node `z` and force `F_z`.
        """

        def __init__(self, name, mech_port, *, f=1e6, mass=1e-6, Q=1e4):
            super().__init__(name)
            self.mass = mass
            self.f = f
            self.Q = Q

            # Add motion and force nodes to mech port.
            # Here we duplicate the already created mechanical
            # nodes in some other connector element
            self._add_port("mech", NodeType.MECHANICAL)
            self.mech._add_node("z", None, mech_port.z)
            self.mech._add_node("F_z", None, mech_port.F_z)
            self._register_node_coupling("F_2_Z", self.mech.F_z, self.mech.z)

        def _on_add(self, model):
            self.mech.z.frequencies = self.frequencies
            self.mech.F_z.frequencies = self.frequencies

        def _get_workspace(self, sim):
            if sim.carrier.any_frequencies_changing:
                # ddb - This case probably needs some more thought on what beatings change
                raise NotImplementedError(
                    "Changing carrier frequencies whilst using a MicroResonant not supported yet."
                )

            if sim.signal:
                refill = (
                    sim.model.fsig.f.is_changing
                    or self.mass.is_changing
                    or self.f.is_changing
                    or self.Q.is_changing
                )
                ws = MicroResonatorWorkspace(self, sim)
                ws.signal.add_fill_function(self.fill, refill)
                ws.Fz_frequencies = sim.signal.signal_frequencies[self.mech.F_z]
                ws.z_frequencies = sim.signal.signal_frequencies[self.mech.z]
                return ws

            return None

        def _couples_frequency(self, ws, connection, frequency_in, frequency_out):
            return True

        @property
        def frequencies(self):
            """Mechanical frequencies this microresonant should model."""
            return (
                self._model.fsig.f.ref,
                self.f.ref + self._model.fsig.f.ref,
                self.f.ref - self._model.fsig.f.ref,
            )

        def fill(self, ws):
            wm = 2 * np.pi * ws.values.f

            Hz = lambda w: 1 / (
                self.mass.value * (wm**2 - w**2 + 1j * w * wm / ws.values.Q)
            )

            for fi in ws.Fz_frequencies.frequencies:
                for fo in ws.z_frequencies.frequencies:
                    if fi.f == fo.f:
                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.F_2_Z_idx,
                            fi.index,
                            fo.index,
                        ) as mat:
                            mat[:] = Hz(2 * np.pi * fo.f)

    kat = finesse.Model()
    kat.parse(
        """
    l l1 P=1
    s sa l1.p1 bs1.p1
    l l2 P=1e3 f=1e6
    s sb l2.p1 bs1.p4
    bs bs1 R=0.5 T=0.5
    s s1 bs1.p3 m1.p1
    m m1 R=1 T=0

    fsig(1)
    sgen sg1 l2.amp.i 1 0

    # ad z m1.mech.z fsig
    ad probe_l m1.p1.o l1.f-fsig
    ad probe_u m1.p1.o l1.f+fsig
    ad pump_l m1.p1.o l2.f-fsig
    ad pump_u m1.p1.o l2.f+fsig
    """
    )
    kat.add(MicroResonator("ures", kat.m1.mech))
    kat.run()



"""Collection of Actions that deal linear time invariant (LTI) modelling tasks."""

from finesse.analysis.runners import (
    run_fsig_sweep,
    run_fsig_sweep2,
    run_fsig_sweep3,
    run_fsig_sweep4,
)
from finesse.solutions import BaseSolution
from finesse.components import DegreeOfFreedom
from finesse.analysis.actions.base import Action, names_to_nodes
import numpy as np
import logging

from finesse.components.node import NodeType
from finesse.exceptions import FinesseException
import collections.abc
from finesse.utilities.misc import deprecation_warning
from typing import List, Tuple

LOGGER = logging.getLogger(__name__)


class FrequencyResponseSolution(BaseSolution):
    """A solution from running a :class:`.FrequencyResponse` action on a model. This
    solution contains the frequency vector and potentially multiple input and output
    transfer function matrix.

    Attributes
    ----------
    f : array_like
        Frequency vector [Hz]
    inputs : array_like
        The input names injected into for this analysis
    outputs : array_like
        The output names read out for this analysis
    out : arrray_like[dtype=np.complex128]
        A matrix of transfer functions for each input to every output over the
        array of frequencies requested. Depending on which frequency response
        action was run will decide what shape this output matrix actually is.
        The shape of out is dependent on the analysis done:

            - FrequencyResponse - (N_f, N_outputs, N_inputs)
            - FrequencyResponse2 - (N_f, N_outputs, N_inputs, N_hom)
            - FrequencyResponse3 - (N_f, N_outputs, N_inputs, N_hom, N_hom)
            - FrequencyResponse4 - (N_f, N_outputs, N_inputs, N_hom)
    type : type
        Type of FrequencyResponse that was used to generate this solution

    Examples
    --------
    Note that the name indexing below is only available when used with the
    :class:`.FrequencyResponse` action, the other frequency-response actions
    must be accessed using the `out` attribute.

    Results from a `FrequencyResponseSolution` can be retrieved in two ways, first
    through the `FrequencyResponseSolution.out` array or by name using
    `[outputs, inputs]`. As an example we will create a fake solution:

    >>> from finesse.analysis.actions.lti import FrequencyResponseSolution
    >>> sol = FrequencyResponseSolution("name")
    >>> sol.inputs = ("A", "B", "C")
    >>> sol.outputs = ("D", "E", "F", "G")
    >>> sol.out = np.random.rand(3, len(sol.outputs), len(sol.inputs))

    The names will map to those provided in the `FrequencyResponse` action you
    called to generate the solution.

    The following will work to select single transfer functions between some input
    and output by name:

    >>> sol["D", "A"] # Select A -> D
    >>> sol["D", "C"] # Select C -> D
    >>> sol["F", "C"] # Select C -> F

    Transfer function matrices can be extracted by providing multiple

    >>> sol["F", ("C", "A")]
    >>> sol[("F", "G"), ("C", "A")]

    Slicing can also be used:

    >>> sol["D", :]   # Select all inputs to "D"
    >>> sol["D", ::2] # Select every other input to "D"
    >>> sol[:, "B"]   # Select "B" to all outputs
    >>> sol[1:, "B"]  # Select "B" to all but the first output
    """

    def outputs_inputs_indices(
        self, outputs: List[str], inputs: List[str], *, reversed: bool = False
    ) -> Tuple[int, int]:
        """Returns the indices to use for selecting certain inputs and outputs from the
        `out` attribute of this solution object.

        Parameters
        ----------
        outputs : List[str]
            List of names of outputs, see `outputs` attribute for options
        inputs : List[str]
            List of names of inputs, see `inputs` attribute for options
        reversed : bool, optional
            Should not be used by the user, this is an internal flag used for
            checking if the reverse case should be tried, ie. inputs specified
            first rather than outputs. This is for backwards compatibility
            before FrequencyResponse was fixed.

        Returns
        -------
        output_index, input_index
            Indices for the outputs and inputs

        Raises
        ------
        KeyError
            Raised when no input or output name can be found.
        """
        out_idx = []
        inp_idx = []

        try:
            out_keys = outputs
            inp_keys = inputs
            out_keys = np.atleast_1d(out_keys)
            inp_keys = np.atleast_1d(inp_keys)

            if type(out_keys) is slice:
                out_idx = out_keys  # just use the slice provided
            else:
                for out_key in out_keys:
                    if isinstance(out_key, slice):
                        out_idx.append(out_key)
                    elif out_key not in self.outputs:
                        raise KeyError(
                            f"This solution does not have an output called '{out_key} {type(out_key)}'. Allowable OUTPUTS={self.outputs}"
                        )
                    else:
                        out_idx.append(self.outputs.index(out_key))
                # single element doesn't need to be an iterable
                if len(out_idx) == 1:
                    out_idx = out_idx[0]

            if type(inp_keys) is slice:
                inp_idx = inp_keys  # just use the slice provided
            else:
                for inp_key in inp_keys:
                    if isinstance(inp_key, slice):
                        inp_idx.append(inp_key)
                    elif inp_key not in self.inputs:
                        raise KeyError(
                            f"This solution does not have an input called '{inp_key} {type(inp_key)}'. Allowable INPUTS={self.inputs}"
                        )
                    else:
                        inp_idx.append(self.inputs.index(inp_key))
                # single element doesn't need to be an iterable
                if len(inp_idx) == 1:
                    inp_idx = inp_idx[0]

            slices = out_idx, inp_idx
            return slices

        except KeyError as ex:
            if reversed:
                return None  # we're trying the reverse, don't keep trying

            # Try and get the transpose [input, output]. eventually this should
            # be deprecated
            result = self.outputs_inputs_indices(inputs, outputs, reversed=True)
            if result is None:
                raise ex  # If the reversed failed then re-raise original exception
            else:
                deprecation_warning(
                    "FrequencyResponseSolution has changed to use [output, input], you seemed to have used [input, output] so returning that.",
                    "3.0",
                )
            return result

    def __getitem__(self, key, *, reversed=False):
        if isinstance(key, str) and self.name == key:
            return self
        elif (
            isinstance(key, (str, bytes))
            or not isinstance(key, collections.abc.Iterable)
            or len(key) != 2
        ):
            raise KeyError(
                """Provide 2 keys [output, input] to select a transfer function
                for indexing this FrequencyResponseSolution, if you want to
                select all of one input or output use a color `:`. Otherwise use the
                `out` attribute to access the underlying data. Shape of `out` is:
                - FrequencyResponse - (N_f, N_outputs, N_inputs)
                - FrequencyResponse2 - (N_f, N_outputs, N_inputs, N_hom)
                - FrequencyResponse3 - (N_f, N_outputs, N_inputs, N_hom, N_hom)
                - FrequencyResponse4 - (N_f, N_outputs, N_inputs, N_hom)
                """
            )
        else:
            o_idx, i_idx = self.outputs_inputs_indices(key[0], key[1])
            return self.out[slice(None), o_idx, i_idx]

    def plot_inputs(self, *inputs, axs=None, max_width=12, show_unity=False, **kwargs):
        """Plot all transfer functions on a NxM grid with a max_width.

        Parameters
        ----------
        inputs*,
            Names of inputs for each subplot
        axs : _type_, optional
            Matplotlib axes to draw on
        max_width : int, optional
            Maximum number of subplots in width
        show_unity : bool, optional
            Plot a line where unity is

        Returns
        -------
        figure, axes
            Matplotlib figure and axes to plot on
        """

        import matplotlib.pyplot as plt
        import numpy as np

        if "show" in kwargs:
            del kwargs["show"]

        if len(inputs) == 0:
            inputs = self.inputs

        inputs = np.atleast_1d(inputs)

        if axs is None:
            # if no axes are given then grab the figure
            # and any axes that are in it
            fig = plt.gcf()
            axs = np.atleast_2d(fig.axes)
        else:
            axs = np.atleast_2d(axs)
            fig = axs[0, 0].figure()

        N = len(inputs)
        W = min(5, max_width / N)
        if np.prod(axs.shape) != N:
            fig, axs = plt.subplots(
                1, N, figsize=(W * N, 3.5), squeeze=False, sharey=True
            )

        if "label" not in kwargs:
            kwargs["label"] = self.outputs

        for i, input in enumerate(inputs):
            axs[0, i].loglog(self.f, abs(self[:, input]), **kwargs)
            axs[0, i].set_xlabel("Frequency [Hz]")
            axs[0, i].set_title(input)
            axs[0, i].legend()
            if show_unity:
                axs[0, i].hlines(
                    1, min(self.f), max(self.f), color="k", ls=":", zorder=-10
                )

        axs[0, 0].set_ylabel("OUTPUT/DOF")
        plt.tight_layout()

        return fig, axs

    plot = plot_inputs  # Default plot option

    def plot_outputs(self, *outputs, axs=None, max_width=12, **kwargs):
        """Plot all transfer functions on a NxM grid with a max_width.

        Parameters
        ----------
        outputs*,
            Names of outputs for each subplot
        axs : _type_, optional
            Matplotlib axes to draw on
        max_width : int, optional
            Maximum number of subplots in width
        show_unity : bool, optional
            Plot a line where unity is

        Returns
        -------
        figure, axes
            Matplotlib figure and axes to plot on
        """

        import matplotlib.pyplot as plt

        if len(outputs) == 0:
            outputs = self.outputs

        outputs = np.atleast_1d(outputs)

        if axs is None:
            N = len(outputs)
            W = min(5, max_width / N)
            fig, axs = plt.subplots(
                1, N, figsize=(W * N, 3.5), squeeze=False, sharey=True
            )
        else:
            fig = plt.gcf()

        if "label" not in kwargs:
            kwargs["label"] = self.inputs

        for i, output in enumerate(outputs):
            axs[0, i].loglog(self.f, abs(self[output, :]), **kwargs)
            axs[0, i].set_xlabel("Frequency [Hz]")
            axs[0, i].set_title(output)
            axs[0, i].legend()

        axs[0, 0].set_ylabel("OUTPUT/DOF")
        plt.tight_layout()

        return fig, axs


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class FrequencyResponse(Action):
    """Computes the frequency response of a signal injected at various nodes to compute
    transfer functions to multiple output nodes. Inputs and outputs should be electrical
    or mechanical nodes. It does this in an efficient way by using the same model and
    solving for multiple RHS input vectors.

    This action does not alter the model state. This action will ignore any currently
    definied signal generator elements in the model.

    Produces an output transfer matrix from each input node to some readout output.
    The shape of the output matrix is:

    [frequencies, inputs, outputs]

    To inject into optical nodes please see :class:`.FrequencyResponse2` and
    :class:`.FrequencyResponse3`. To readout optical nodes please see
    :class:`.FrequencyResponse3` and :class:`.FrequencyResponse4`.

    Parameters
    ----------
    f : array, double
        Frequencies to compute the transfer functions over
    inputs : iterable[str or Element]
        Mechanical or electrical node to inject signal at
    outputs : iterable[str or Element]
        Mechanical or electrical nodes to measure output at
    open_loop : bool, optional
        Computes open loop transfer functions if the system has closed
    name : str, optional
        Solution name

    Examples
    --------
    Here we measure a set of transfer functions from DARM and CARM
    to four readouts for a particular `model`,

    >>> sol = model.run(FrequencyResponse(np.geomspace(0.1, 50000, 100),
    ...         ('DARM', 'CARM'),
    ...         ('AS.DC', 'AS45.I', 'AS45.Q', 'REFL9.I'),
    ... ))

    Single inputs and outputs can also be specified

    >>> model.run(FrequencyResponse(np.geomspace(0.1, 50000, 100), 'DARM', AS.DC'))

    The transfer functions can then be accessed like a 2D array by name,
    the ordering of inputs to outputs does not matter.

    >>> sol['DARM'] # DARM to all outputs
    >>> sol['DARM', 'AS.DC'] # DARM to AS.DC
    >>> sol['DARM', ('AS.DC', 'AS45.I')]
    >>> sol['AS.DC'] # All inputs to AS.DC readout
    """

    def __init__(
        self, f, inputs, outputs, *, open_loop=False, name="frequency_response"
    ):
        super().__init__(name)
        inputs = np.atleast_1d(inputs)
        outputs = np.atleast_1d(outputs)
        if f is None:
            raise FinesseException("A frequency vector must be provided")

        try:
            self.f = np.array(f, dtype=np.float64, copy=True)
        except TypeError:
            # If the f is a symbol...
            self.f = np.array(f.eval(), dtype=np.float64, copy=True)
        if self.f.size == 0:
            raise FinesseException("Frequency vector has size 0")
        if any(self.f <= 0):
            raise FinesseException(
                "Frequency vector must contain values greater than 0"
            )

        def process(x, input):
            if isinstance(x, DegreeOfFreedom):
                if input:
                    return x.AC.i.full_name
                else:
                    return x.AC.o.full_name
            elif isinstance(x, (str, np.str_)):
                return x
            else:  # Try and get full_name
                return x.full_name

        self.inputs = list(process(i, True) for i in inputs)
        self.outputs = list(process(o, False) for o in outputs)
        self.open_loop = open_loop

    def _do(self, state, fsig_independant_outputs=None, fsig_dependant_outputs=None):
        input_node_indices = np.zeros(len(self.inputs), dtype=np.dtype("long"))
        output_node_indices = np.zeros(len(self.outputs), dtype=np.dtype("long"))

        # some signals will need to be scaled
        input_scaling = np.ones(len(self.inputs), dtype=float)
        output_scaling = np.ones(len(self.outputs), dtype=float)

        for i, node in enumerate(
            names_to_nodes(state.model, self.inputs, default_hints=("input",))
        ):
            if node.type is NodeType.OPTICAL:
                raise FinesseException(
                    f"Optical nodes ({node}) cannot be used with the FrequencyResponse action"
                )
            else:
                # set scaling for mechanical input signals
                if node.type is NodeType.MECHANICAL:
                    input_scaling[i] /= state.sim.model_settings.x_scale

                input_node_indices[i] = state.sim.signal.node_id(node)

        for i, node in enumerate(
            names_to_nodes(state.model, self.outputs, default_hints=("output",))
        ):
            if node.type is NodeType.OPTICAL:
                raise FinesseException(
                    f"Optical nodes ({node}) cannot be used with the FrequencyResponse action"
                )
            else:
                # set scaling for mechanical output signals
                if node.type is NodeType.MECHANICAL:
                    output_scaling[i] *= state.sim.model_settings.x_scale

                output_node_indices[i] = state.sim.signal.node_id(node)

        sol = FrequencyResponseSolution(self.name)
        sol.f = self.f
        sol.inputs = self.inputs
        sol.outputs = self.outputs
        state.sim.run_carrier()
        rtn = run_fsig_sweep(
            state.sim,
            self.f,
            input_node_indices,
            output_node_indices,
            input_scaling,
            output_scaling,
            None,
            self.open_loop,
            (
                tuple(fsig_independant_outputs)
                if fsig_independant_outputs is not None
                else None
            ),
            (
                tuple(fsig_dependant_outputs)
                if fsig_dependant_outputs is not None
                else None
            ),
        )
        if (fsig_dependant_outputs is not None) or (
            fsig_independant_outputs is not None
        ):
            sol.out = rtn[0]
            sol.extra_outputs = rtn[1]
        else:
            sol.out = rtn

        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].append("fsig.f")
        memo["input_nodes"].extend((_, ("input",)) for _ in self.inputs)
        memo["output_nodes"].extend((_, ("output",)) for _ in self.outputs)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class FrequencyResponse2(Action):
    """Computes the frequency response of a signal injected at an optical port at a
    particular optical frequency. This differs from :class:`.FrequencyResponse` in the
    way the inputs and outputs are prescribed. For :class:`.FrequencyResponse2` you
    specify optical input nodes and a signal output node.

    This action does not alter the model state. This action will ignore any currently
    definied signal generator elements in the model.

    Produces an output transfer matrix from each HOM at a particular frequency and
    optical node to some readout output. The shape of the output matrix is:

    [frequencies, outputs, inputs, HOMs]

    It should be noted that when exciting a lower signal sideband frequency it will
    actually return the operator for propagating the conjugate of the lower sideband.
    This is because FINESSE is internally solving for the conjugate of the lower sideband
    to linearise non-linear optical effects.

    To inject into mechanical and electrical nodes please see :class:`.FrequencyResponse`
    and :class:`.FrequencyResponse4`. To readout optical nodes please see
    :class:`.FrequencyResponse3` and :class:`.FrequencyResponse4`.

    Parameters
    ----------
    f : array, double
        Frequencies to compute the transfer functions over
    inputs : iterable[tuple[str or Node, Frequency]]
        Optical node and frequency tuple to inject at. A symbolic refence to the
        model's fsig.f parameter should always be used when defining a frequency to
        look at.
    outputs : iterable[str or Element]
        Mechanical or electrical (signal)nodes to measure output to
    name : str, optional
        Solution name

    Examples
    --------
    It is advisable to use always use a reference to the symbolic reference to
    the signal frequency `model.fsig.f.ref` instead of a fixed number incase it
    changes. This action will look for an initial frequency bin of X Hz to track
    during the frequency response analysis. A symbolic reference will always
    ensure the right bin is used, in cases such as looking at RF signal sidebands,
    `10e6+model.fsig.f.ref` and `10e6-model.fsig.f.ref` will always look at the
    upper and lower signal sideband around the +10MHz sideband.

    >>> import finesse
    >>> from finesse.analysis.actions import FrequencyResponse2
    >>> model = finesse.script.parse('''
    ... l l1
    ... bs bs1 R=1 T=0 xbeta=1e-6 ybeta=1e-9
    ... readout_dc A
    ... link(l1, bs1, A)
    ... fsig(1)
    ... modes(maxtem=1)
    ... gauss g1 l1.p1.o w=1m Rc=inf
    ... ''')
    >>> sol = model.run(
    ...     FrequencyResponse2(
    ...         [1, 10, 100],
    ...         [
    ...             ('bs1.p2.o', +model.fsig.f.ref),
    ...             ('bs1.p2.o', -model.fsig.f.ref)
    ...         ],
    ...         ['A.DC']
    ...     )
    ... )
    """

    def __init__(self, f, inputs, outputs, *, name="frequency_response2"):
        super().__init__(name)

        if f is None:
            raise FinesseException("A frequency vector must be provided")
        try:
            self.f = np.array(f, dtype=np.float64, copy=True)
        except TypeError:
            # If the f is a symbol...
            self.f = np.array(f.eval(), dtype=np.float64, copy=True)
        if self.f.size == 0:
            raise FinesseException("Frequency vector has size 0")
        if any(self.f <= 0):
            raise FinesseException(
                "Frequency vector must contain values greater than 0"
            )

        self.inputs = inputs
        self.outputs = outputs

        self.input_nodes = []
        self.input_freqs = []
        for node, freq in inputs:
            self.input_nodes.append(node)
            self.input_freqs.append(freq)

        self.output_nodes = []
        for node in outputs:
            self.output_nodes.append(node)

        def process_node(x, input):
            if isinstance(x, DegreeOfFreedom):
                if input:
                    return x.AC.i.full_name
                else:
                    return x.AC.o.full_name
            elif isinstance(x, (str, np.str_)):
                return x
            else:  # Try and get full_name
                return x.full_name

        self.input_nodes = list(process_node(i, True) for i in self.input_nodes)
        self.output_nodes = list(process_node(o, False) for o in self.output_nodes)

    def _do(self, state, fsig_independant_outputs=None, fsig_dependant_outputs=None):
        input_node_indices = np.zeros(len(self.input_nodes), dtype=np.dtype("long"))
        input_freq_indices = np.zeros(len(self.input_nodes), dtype=np.dtype("long"))
        output_node_indices = np.zeros(len(self.output_nodes), dtype=np.dtype("long"))

        # some signals will need to be scaled
        input_scaling = np.ones(len(self.input_nodes), dtype=float)
        output_scaling = np.ones(len(self.output_nodes), dtype=float)

        for i, (node, freq) in enumerate(
            zip(
                names_to_nodes(state.model, self.input_nodes, default_hints=("input",)),
                self.input_freqs,
            )
        ):
            freq_obj = state.sim.signal.get_frequency_object(freq, node)
            input_freq_indices[i] = freq_obj.index

            if node.type is NodeType.OPTICAL:
                input_node_indices[i] = state.sim.signal.node_id(node)
            else:
                if input_freq_indices[i] != 0:
                    raise FinesseException(
                        f"Input frequency for {node} should be the signal frequency"
                    )
                # set scaling for mechanical input signals
                if node.type is NodeType.MECHANICAL:
                    input_scaling[i] /= state.sim.model_settings.x_scale
                input_node_indices[i] = state.sim.signal.node_id(node)

        for i, node in enumerate(
            names_to_nodes(state.model, self.output_nodes, default_hints=("output",))
        ):
            if node.type is NodeType.OPTICAL:
                raise FinesseException(
                    f"Optical nodes ({node}) cannot be used with the FrequencyResponse2 action"
                )
            else:
                # set scaling for mechanical output signals
                if node.type is NodeType.MECHANICAL:
                    output_scaling[i] *= state.sim.model_settings.x_scale
                output_node_indices[i] = state.sim.signal.node_id(node)

        sol = FrequencyResponseSolution(self.name)
        sol.type = FrequencyResponse2
        sol.f = self.f
        sol.inputs = self.inputs
        sol.outputs = self.outputs
        state.sim.run_carrier()
        rtn = run_fsig_sweep2(
            state.sim,
            self.f,
            input_node_indices,
            input_freq_indices,
            output_node_indices,
            input_scaling,
            output_scaling,
            None,
            (
                tuple(fsig_independant_outputs)
                if fsig_independant_outputs is not None
                else None
            ),
            (
                tuple(fsig_dependant_outputs)
                if fsig_dependant_outputs is not None
                else None
            ),
        )
        if (fsig_dependant_outputs is not None) or (
            fsig_independant_outputs is not None
        ):
            sol.out = rtn[0]
            sol.extra_outputs = rtn[1]
        else:
            sol.out = rtn

        return sol

    def _requests(self, model, memo, first=True):
        for freq in self.input_freqs:
            try:
                if model.fsig.f.ref not in freq.parameters():
                    raise IndexError()
            except (AttributeError, IndexError):  # catch if freq not a symbol
                raise FinesseException(
                    f"{self} requires frequencies to be specified as a symbolic expression which must include `model.fsig.f.ref`, not {freq}."
                )

        memo["changing_parameters"].append("fsig.f")
        memo["input_nodes"].extend((_, ("input",)) for _ in self.input_nodes)
        memo["output_nodes"].extend((_, ("output",)) for _ in self.output_nodes)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class FrequencyResponse3(Action):
    """Computes the frequency response of a signal injected at an optical port at a
    particular optical frequency. This differs from :class:`.FrequencyResponse` in the
    way the inputs and outputs are prescribed. For :class:`.FrequencyResponse3` you
    specify optical input nodes and optical output nodes.

    This action does not alter the model state. This action will ignore any currently
    definied signal generator elements in the model.

    Produces an output transfer matrix from each HOM at a particular frequency and
    optical node to some other optical node and frequency. The shape of the output
    matrix is:

    [frequencies, outputs, inputs, HOMs, HOMs]

    It should be noted that when exciting a lower signal sideband frequency it will
    actually return the operator for propagating the conjugate of the lower sideband.
    This is because FINESSE is internally solving for the conjugate of the lower sideband
    to linearise non-linear optical effects.

    To inject into mechanical and electrical nodes please see :class:`.FrequencyResponse`
    and :class:`.FrequencyResponse4`. To readout mechanical and electrical nodes
    please see :class:`.FrequencyResponse` and :class:`.FrequencyResponse2`.

    Parameters
    ----------
    f : array, double
        Frequencies to compute the transfer functions over
    inputs : iterable[tuple[str or Node, Frequency]]
        Optical node and frequency tuple to inject at. A symbolic reference to the
        model's fsig.f parameter should always be used when defining a frequency to
        look at.
    outputs : iterable[tuple[str or Node, Frequency]]
        Optical node and frequency tuple to measure output at. A symbolic reference to the
        model's fsig.f parameter should always be used when defining a frequency to
        look at.
    name : str, optional
        Solution name

    Examples
    --------
    It is advisable to use always use a reference to the symbolic reference to
    the signal frequency `model.fsig.f.ref` instead of a fixed number incase it
    changes. This action will look for an initial frequency bin of X Hz to track
    during the frequency response analysis. A symbolic reference will always
    ensure the right bin is used, in cases such as looking at RF signal sidebands,
    `10e6+model.fsig.f.ref` and `10e6-model.fsig.f.ref` will always look at the
    upper and lower signal sideband around the +10MHz sideband.

    >>> import finesse
    >>> from finesse.analysis.actions import FrequencyResponse3
    >>> model = finesse.script.parse('''
    ... l l1
    ... bs bs1 R=1 T=0 xbeta=1e-6 ybeta=1e-9
    ... readout_dc A
    ... link(l1, bs1, A)
    ... fsig(1)
    ... modes(maxtem=1)
    ... gauss g1 l1.p1.o w=1m Rc=inf
    ... ''')
    >>> sol = model.run(
    ...     FrequencyResponse3(
    ...         [1, 10, 100],
    ...         [
    ...             ('bs1.p2.o', +model.fsig.f.ref),
    ...             ('bs1.p2.o', -model.fsig.f.ref)
    ...         ],
    ...         [
    ...             ('A.p1.i', +model.fsig.f.ref),
    ...             ('A.p1.i', -model.fsig.f.ref)
    ...         ]
    ...     )
    ... )
    """

    def __init__(self, f, inputs, outputs, *, name="frequency_response3"):
        super().__init__(name)

        if f is None:
            raise FinesseException("A frequency vector must be provided")
        try:
            self.f = np.array(f, dtype=np.float64, copy=True)
        except TypeError:
            # If the f is a symbol...
            self.f = np.array(f.eval(), dtype=np.float64, copy=True)
        if self.f.size == 0:
            raise FinesseException("Frequency vector has size 0")
        if any(self.f <= 0):
            raise FinesseException(
                "Frequency vector must contain values greater than 0"
            )

        self.inputs = inputs
        self.outputs = outputs

        self.input_nodes = []
        self.input_freqs = []
        for node, freq in inputs:
            self.input_nodes.append(node)
            self.input_freqs.append(freq)

        self.output_nodes = []
        self.output_freqs = []
        for node, freq in outputs:
            self.output_nodes.append(node)
            self.output_freqs.append(freq)

        def process_node(x, input):
            if isinstance(x, DegreeOfFreedom):
                if input:
                    return x.AC.i.full_name
                else:
                    return x.AC.o.full_name
            elif isinstance(x, (str, np.str_)):
                return x
            else:  # Try and get full_name
                return x.full_name

        self.input_nodes = list(process_node(i, True) for i in self.input_nodes)
        self.output_nodes = list(process_node(o, False) for o in self.output_nodes)

    def _do(self, state, fsig_independant_outputs=None, fsig_dependant_outputs=None):
        input_node_indices = np.zeros(len(self.input_nodes), dtype=np.dtype("long"))
        input_freq_indices = np.zeros(len(self.input_nodes), dtype=np.dtype("long"))
        output_node_indices = np.zeros(len(self.output_nodes), dtype=np.dtype("long"))
        output_freq_indices = np.zeros(len(self.output_nodes), dtype=np.dtype("long"))

        # some signals will need to be scaled
        input_scaling = np.ones(len(self.input_nodes), dtype=float)
        output_scaling = np.ones(len(self.output_nodes), dtype=float)

        for i, (node, freq) in enumerate(
            zip(
                names_to_nodes(state.model, self.input_nodes, default_hints=("input",)),
                self.input_freqs,
            )
        ):
            if node.type is NodeType.OPTICAL:
                freq_obj = state.sim.signal.get_frequency_object(freq, node)
                input_freq_indices[i] = freq_obj.index
                input_node_indices[i] = state.sim.signal.node_id(node)
            else:
                raise FinesseException(
                    f"Optical nodes ({node}) must be used with the FrequencyResponse3 action"
                )

        for i, (node, freq) in enumerate(
            zip(
                names_to_nodes(
                    state.model, self.output_nodes, default_hints=("output",)
                ),
                self.output_freqs,
            )
        ):
            if node.type is NodeType.OPTICAL:
                freq_obj = state.sim.signal.get_frequency_object(freq, node)
                output_freq_indices[i] = freq_obj.index
                output_node_indices[i] = state.sim.signal.node_id(node)
            else:
                raise FinesseException(
                    f"Optical nodes ({node}) must be used with the FrequencyResponse3 action"
                )

        sol = FrequencyResponseSolution(self.name)
        sol.type = FrequencyResponse3
        sol.f = self.f
        sol.inputs = self.inputs
        sol.outputs = self.outputs
        state.sim.run_carrier()
        rtn = run_fsig_sweep3(
            state.sim,
            self.f,
            input_node_indices,
            input_freq_indices,
            output_node_indices,
            output_freq_indices,
            input_scaling,
            output_scaling,
            None,
            (
                tuple(fsig_independant_outputs)
                if fsig_independant_outputs is not None
                else None
            ),
            (
                tuple(fsig_dependant_outputs)
                if fsig_dependant_outputs is not None
                else None
            ),
        )
        if (fsig_dependant_outputs is not None) or (
            fsig_independant_outputs is not None
        ):
            sol.out = rtn[0]
            sol.extra_outputs = rtn[1]
        else:
            sol.out = rtn

        return sol

    def _requests(self, model, memo, first=True):
        for flist in [self.input_freqs, self.output_freqs]:
            for freq in flist:
                try:
                    if model.fsig.f.ref not in freq.parameters():
                        raise IndexError()
                except (AttributeError, IndexError):  # catch if freq not a symbol
                    raise FinesseException(
                        f"{self} requires frequencies to be specified as a symbolic expression which must include `model.fsig.f.ref`, not {repr(freq)}."
                    )

        memo["changing_parameters"].append("fsig.f")
        memo["input_nodes"].extend((_, ("input",)) for _ in self.input_nodes)
        memo["output_nodes"].extend((_, ("output",)) for _ in self.output_nodes)


class FrequencyResponse4(Action):
    """Computes the frequency response of a signal injected at an electrical or
    mechanical port. This differs from :class:`.FrequencyResponse` in the way the inputs
    and outputs are prescribed. For :class:`.FrequencyResponse4` you specify signal
    input nodes and optical output nodes.

    This action does not alter the model state. This action will ignore any currently
    defined signal generator elements in the model.

    Produces an output transfer matrix from each signal node to each HOM at a
    particular frequency and optical node. The shape of the output
    matrix is:

    [frequencies, outputs, inputs, HOMs]

    It should be noted that when exciting a lower signal sideband frequency it will
    actually return the operator for propagating the conjugate of the lower sideband.
    This is because FINESSE is internally solving for the conjugate of the lower sideband
    to linearise non-linear optical effects.

    To inject into optical nodes please see :class:`.FrequencyResponse2` and
    :class:`.FrequencyResponse3`. To readout mechanical and electrical nodes
    please see :class:`.FrequencyResponse` and :class:`.FrequencyResponse2`.

    Parameters
    ----------
    f : array, double
        Frequencies to compute the transfer functions over
    inputs : iterable[str or Element]
        Mechanical or electrical node to inject signal at
    outputs : iterable[tuple[str or Node, Frequency]]
        Optical node and frequency tuple to measure output at. A symbolic reference to the
        model's fsig.f parameter should always be used when defining a frequency to
        look at.
    name : str, optional
        Solution name

    Examples
    --------
    It is advisable to use always use a reference to the symbolic reference to
    the signal frequency `model.fsig.f.ref` instead of a fixed number incase it
    changes. This action will look for an initial frequency bin of X Hz to track
    during the frequency response analysis. A symbolic reference will always
    ensure the right bin is used, in cases such as looking at RF signal sidebands,
    `10e6+model.fsig.f.ref` and `10e6-model.fsig.f.ref` will always look at the
    upper and lower signal sideband around the +10MHz sideband.

    >>> sol = model.run(
    ...     FrequencyResponse(
    ...         [1, 10, 100],
    ...         [model.ETM.mech.z],
    ...         [
    ...             (model.ITM.p2.o, +model.fsig.f.ref),
    ...             (model.ITM.p2.o, -model.fsig.f.ref)
    ...         ]
    ...     )
    ... )
    """

    def __init__(self, f, inputs, outputs, *, name="frequency_response4"):
        super().__init__(name)

        if f is None:
            raise FinesseException("A frequency vector must be provided")
        try:
            self.f = np.array(f, dtype=np.float64, copy=True)
        except TypeError:
            # If the f is a symbol...
            self.f = np.array(f.eval(), dtype=np.float64, copy=True)
        if self.f.size == 0:
            raise FinesseException("Frequency vector has size 0")
        if any(self.f <= 0):
            raise FinesseException(
                f"Frequency vector: {f} must contain values greater than 0"
            )

        self.inputs = inputs
        self.outputs = outputs

        self.input_nodes = []
        for node in inputs:
            self.input_nodes.append(node)

        self.output_nodes = []
        self.output_freqs = []
        for node, freq in outputs:
            self.output_nodes.append(node)
            self.output_freqs.append(freq)

        def process_node(x, input):
            if isinstance(x, DegreeOfFreedom):
                if input:
                    return x.AC.i.full_name
                else:
                    return x.AC.o.full_name
            elif isinstance(x, (str, np.str_)):
                return x
            else:  # Try and get full_name
                return x.full_name

        self.input_nodes = list(process_node(i, True) for i in self.input_nodes)
        self.output_nodes = list(process_node(o, False) for o in self.output_nodes)

    def _do(self, state, fsig_independant_outputs=None, fsig_dependant_outputs=None):
        input_node_indices = np.zeros(len(self.input_nodes), dtype=np.dtype("long"))
        output_node_indices = np.zeros(len(self.output_nodes), dtype=np.dtype("long"))
        output_freq_indices = np.zeros(len(self.output_nodes), dtype=np.dtype("long"))

        # some signals will need to be scaled
        input_scaling = np.ones(len(self.input_nodes), dtype=float)
        output_scaling = np.ones(len(self.output_nodes), dtype=float)

        for i, node in enumerate(
            names_to_nodes(state.model, self.inputs, default_hints=("input",))
        ):
            if node.type is NodeType.OPTICAL:
                raise FinesseException(
                    f"This optical node ({node}) cannot be used with the FrequencyResponse4 action make sure you're using the correct input and outputs."
                )
            else:
                # set scaling for mechanical input signals
                if node.type is NodeType.MECHANICAL:
                    input_scaling[i] /= state.sim.model_settings.x_scale

                input_node_indices[i] = state.sim.signal.node_id(node)

        for i, (node, freq) in enumerate(
            zip(
                names_to_nodes(
                    state.model, self.output_nodes, default_hints=("output",)
                ),
                self.output_freqs,
            )
        ):
            if node.type is NodeType.OPTICAL:
                freq_obj = state.sim.signal.get_frequency_object(freq, node)
                output_freq_indices[i] = freq_obj.index
                output_node_indices[i] = state.sim.signal.node_id(node)
            else:
                raise FinesseException(
                    f"Optical nodes ({node}) must be used with the FrequencyResponse4 action"
                )

        sol = FrequencyResponseSolution(self.name)
        sol.type = FrequencyResponse4
        sol.f = self.f
        sol.inputs = self.inputs
        sol.outputs = self.outputs
        state.sim.run_carrier()
        rtn = run_fsig_sweep4(
            state.sim,
            self.f,
            input_node_indices,
            output_node_indices,
            output_freq_indices,
            input_scaling,
            output_scaling,
            None,
            (
                tuple(fsig_independant_outputs)
                if fsig_independant_outputs is not None
                else None
            ),
            (
                tuple(fsig_dependant_outputs)
                if fsig_dependant_outputs is not None
                else None
            ),
        )
        if (fsig_dependant_outputs is not None) or (
            fsig_independant_outputs is not None
        ):
            sol.out = rtn[0]
            sol.extra_outputs = rtn[1]
        else:
            sol.out = rtn

        return sol

    def _requests(self, model, memo, first=True):
        for freq in self.output_freqs:
            try:
                if model.fsig.f.ref not in freq.parameters():
                    raise IndexError()
            except (AttributeError, IndexError):  # catch if freq not a symbol
                raise FinesseException(
                    f"{self} requires frequencies to be specified as a symbolic expression which must include `model.fsig.f.ref`, not {repr(freq)}."
                )

        memo["changing_parameters"].append("fsig.f")
        memo["input_nodes"].extend((_, ("input",)) for _ in self.input_nodes)
        memo["output_nodes"].extend((_, ("output",)) for _ in self.output_nodes)



"""Axes Actions such as xaxis, noaxais, etc."""

import re
from typing import Literal

from finesse.analysis.actions.base import Action
from finesse.parameter import Parameter, ParameterRef

from .sweep import Sweep, get_sweep_array


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Noxaxis(Sweep):
    """The default Action for a model when no action is specified. Solves the model
    without changing any parameters. Will return a solution with a single value
    for every detector in the model.

    Parameters
    ----------
    pre_step : Action | None, optional
        An action to perform before the Noxaxis is run, by default None
    post_step : Action | None, optional
        An action to perform after the Noxaxis is run, by default None
    name : str, optional
        Name of the action, used to find the solution in the final output, by
        default "noxaxis"
    """

    def __init__(
        self,
        pre_step: Action | None = None,
        post_step: Action | None = None,
        name: str = "noxaxis",
    ):
        super().__init__(name=name, pre_step=pre_step, post_step=post_step)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class XNaxis(Sweep):
    """N-dimensional version of Xaxis allows for multi-dimensional parameter sweep
    of arbitrary dimension N. You need to provide a set of 'parameter, mode, start,
    stop, steps' arguments for every dimension you want to sweep over.

    Parameters
    ----------
    relative : bool, optional
        Whether to apply the changes to the parameter value on top of the current
        value or ignore the current value of the parameter, by default False, by
        default False
    pre_step : Action | None, optional
        An action to perform before each step is computed, by default None
    post_step : Action | None, optional
        An action to perform after each step is computed, by default None
    name : str, optional
        Name of the action, used to find the solution in the final output, by
        default "XNaxis"
    """

    def __init__(
        self,
        *args,
        relative: bool = False,
        pre_step: Action | None = None,
        post_step: Action | None = None,
        name: str = "XNaxis",
    ):
        if len(args) % 5 != 0:
            raise Exception(
                f"XNaxis arguments must come in groups of five: parameter, mode, start, stop, steps not {args}"
            )
        self.relative = relative
        self.N = len(args) // 5

        if self.N == 0:
            raise Exception("XNaxis requires at least one axis to be specified")
        # Here we map the XNaxis arguments to the Sweep inputs
        self.__set_args = args
        new_args = []

        for i in range(0, len(args), 5):
            new_args.append(args[i + 0])
            new_args.append(
                get_sweep_array(args[i + 2], args[i + 3], args[i + 4], args[i + 1])
            )
            new_args.append(relative)
        super().__init__(*new_args, pre_step=pre_step, post_step=post_step, name=name)

    def __getattr__(self, key):
        res = re.match("(parameter|mode|start|stop|steps)([0-9]*)", key)
        if res is None:
            super().__getattribute__(key)
        else:
            grp = res.groups()
            N = 1 if grp[1] == "" else int(grp[1])
            if N == 0:
                raise Exception("Specify an axes greater than 0")
            if N > self.N:
                raise Exception(f"This xaxis does not have {N} axes")
            idx = 5 * (N - 1)
            if grp[0] == "parameter":
                return self.__set_args[idx + 0]
            elif grp[0] == "mode":
                return self.__set_args[idx + 1]
            elif grp[0] == "start":
                return self.__set_args[idx + 2]
            elif grp[0] == "stop":
                return self.__set_args[idx + 3]
            elif grp[0] == "steps":
                return self.__set_args[idx + 4]


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Xaxis(XNaxis):
    """Sweeps a parameter and solves the model for every value in the sweep. Will
    a solution with an array of values for every detector in the model.

    Parameters
    ----------
    parameter : Parameter | ParameterRef | str
        Parameter to sweep over.
    mode : Literal['lin', 'log']
        Whether to sweep linearly or logarithmically
    start : float
        Starting value of the parameter sweep
    stop : float
        Final value of the parameter sweep.
    steps : int
        Number of steps in the sweep. Solution arrays will contain
        'steps+1' values
    relative : bool, optional
        Whether to apply the changes to the parameter value on top of the current
        value or ignore the current value of the parameter, by default False
    pre_step : Action | None, optional
        An action to perform before each step is computed, by default None
    post_step : Action | None, optional
        An action to perform after each step is computed, by default None
    name : str, optional
        Name of the action, used to find the solution in the final output, by
        default "xaxis"
    """

    def __init__(
        self,
        parameter: Parameter | ParameterRef | str,
        mode: Literal["lin", "log"],
        start: float,
        stop: float,
        steps: int,
        relative: bool = False,
        *,
        pre_step: Action | None = None,
        post_step: Action | None = None,
        name: str = "xaxis",
    ):
        super().__init__(
            parameter,
            mode,
            start,
            stop,
            steps,
            relative=relative,
            pre_step=pre_step,
            post_step=post_step,
            name=name,
        )


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class X2axis(XNaxis):
    """Two-dimensional version of the Xaxis action. Will solve the model for every
    point on a grid of values defined by the two sets of sweep arguments. Returns a
    solution with a two-dimensional array for every detector in the model.

    Significantly faster than calling Xaxis in a Python for loop.

    Parameters
    ----------
    parameter1 : Parameter | ParameterRef | str
        Parameter to sweep over.
    mode1 : Literal['lin', 'log']
        Whether to sweep linearly or logarithmically
    start1 : float
        Starting value of the parameter sweep
    stop1 : float
        Final value of the parameter sweep.
    steps1 : int
        Number of steps in the sweep. Solution arrays will contain
        'steps+1' values
    parameter2 : Parameter | ParameterRef | str
        Parameter to sweep over.
    mode2 : Literal['lin', 'log']
        Whether to sweep linearly or logarithmically
    start2 : float
        Starting value of the parameter sweep
    stop2 : float
        Final value of the parameter sweep.
    steps2 : int
        Number of steps in the sweep. Solution arrays will contain
        'steps+1' values
    relative : bool, optional
        Whether to apply the changes to the parameter value on top of the current
        value or ignore the current value of the parameter, by default False, by
        default False
    pre_step : Action | None, optional
        An action to perform before each step is computed, by default None
    post_step : Action | None, optional
        An action to perform after each step is computed, by default None
    name : str, optional
        Name of the action, used to find the solution in the final output, by
        default "x2axis"
    """

    def __init__(
        self,
        parameter1: Parameter | ParameterRef | str,
        mode1: Literal["lin", "log"],
        start1: float,
        stop1: float,
        steps1: int,
        parameter2: Parameter | ParameterRef | str,
        mode2: Literal["lin", "log"],
        start2: float,
        stop2: float,
        steps2: int,
        relative: bool = False,
        *,
        pre_step: Action | None = None,
        post_step: Action | None = None,
        name: str = "x2axis",
    ):
        super().__init__(
            parameter1,
            mode1,
            start1,
            stop1,
            steps1,
            parameter2,
            mode2,
            start2,
            stop2,
            steps2,
            relative=relative,
            pre_step=pre_step,
            post_step=post_step,
            name=name,
        )


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class X3axis(XNaxis):
    """Three-dimensional version of the Xaxis action. Will solve the model for every
    point on a cube of values defined by the three sets of sweep arguments. Returns
    a solution with a three-dimensional array for every detector in the model.

    Significantly faster than calling Xaxis in a nested Python for loop.

    Parameters
    ----------
    parameter1 : Parameter | ParameterRef | str
        Parameter to sweep over.
    mode1 : Literal['lin', 'log']
        Whether to sweep linearly or logarithmically
    start1 : float
        Starting value of the parameter sweep
    stop1 : float
        Final value of the parameter sweep.
    steps1 : int
        Number of steps in the sweep. Solution arrays will contain
        'steps+1' values
    parameter2 : Parameter | ParameterRef | str
        Parameter to sweep over.
    mode2 : Literal['lin', 'log']
        Whether to sweep linearly or logarithmically
    start2 : float
        Starting value of the parameter sweep
    stop2 : float
        Final value of the parameter sweep.
    steps2 : int
        Number of steps in the sweep. Solution arrays will contain
        'steps+1' values
    parameter3 : Parameter | ParameterRef | str
        Parameter to sweep over.
    mode3 : Literal['lin', 'log']
        Whether to sweep linearly or logarithmically
    start3 : float
        Starting value of the parameter sweep
    stop3 : float
        Final value of the parameter sweep.
    steps3 : int
        Number of steps in the sweep. Solution arrays will contain
        'steps+1' values
    relative : bool, optional
        Whether to apply the changes to the parameter value on top of the current
        value or ignore the current value of the parameter, by default False, by
        default False
    pre_step : Action | None, optional
        An action to perform before each step is computed, by default None
    post_step : Action | None, optional
        An action to perform after each step is computed, by default None
    name : str, optional
        Name of the action, used to find the solution in the final output, by
        default "x3axis"
    """

    def __init__(
        self,
        parameter1: Parameter | ParameterRef | str,
        mode1: Literal["lin", "log"],
        start1: float,
        stop1: float,
        steps1: int,
        parameter2: Parameter | ParameterRef | str,
        mode2: Literal["lin", "log"],
        start2: float,
        stop2: float,
        steps2: int,
        parameter3: Parameter | ParameterRef | str,
        mode3: Literal["lin", "log"],
        start3: float,
        stop3: float,
        steps3: int,
        relative: bool = False,
        *,
        pre_step: Action | None = None,
        post_step: Action | None = None,
        name="x3axis",
    ):
        super().__init__(
            parameter1,
            mode1,
            start1,
            stop1,
            steps1,
            parameter2,
            mode2,
            start2,
            stop2,
            steps2,
            parameter3,
            mode3,
            start3,
            stop3,
            steps3,
            relative=relative,
            pre_step=pre_step,
            post_step=post_step,
            name=name,
        )



#cython: unraisable_tracebacks=True
"""Base solution interface.

Solution classes contain the output from a Finesse simulation, and convenience methods
for accessing, plotting and serialising that output.

Solutions intentionally do not contain references to the model that produced its
results. This is so that the solution can be serialised without requiring the model that
produced it itself be serialisable.
"""
import logging
from collections.abc import Set, Iterable
from itertools import zip_longest
from functools import reduce
import numpy as np
from finesse.tree cimport TreeNode
from finesse.solutions.array import ArraySolutionSet
from finesse.solutions.array import ArraySolution
from finesse.exceptions import FinesseException

LOGGER = logging.getLogger(__name__)


cdef class ParameterChangingTreeNode(TreeNode):
    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.parameters_changing = ()

    def get_all_parameters_changing(self):
        return reduce(
            lambda a, b: (*a, *b.parameters_changing),
            self.get_all_children(),
            (*self.parameters_changing,),
        )


class SolutionSet(Set):
    """
    A SolutionSet is a collection of solution objects that have been generated
    from running a model. When nested analyses have been used then you must then
    extract each of the nested solutions. The SolutionSet object makes this easier
    by collecting a variety of Solutions together so you can select common nested
    solutions within them, or extract some common attribute.

    Take some overly simplistic example here where we have swept some variable
    and performed some other analysis at each step.

    >>> import finesse
    >>> model = finesse.Model()
    >>> model.parse('''
    ... l l1 P=1
    ... pd P l1.p1.o
    ... xaxis(l1.P, lin, 0, 1, 2,
    ...     pre_step=series(
    ...         minimize(P, l1.P)
    ...     )
    ... )
    ... ''')
    >>>
    >>> sol = model.run()
    >>> print(sol)
    - Solution Tree
    ● xaxis - ArraySolution
    ╰──○ pre_step
       ├──○ series
       │  ╰──○ minimize - OptimizeSolution
       ├──○ series
       │  ╰──○ minimize - OptimizeSolution
       ╰──○ series
           ╰──○ minimize - OptimizeSolution

    Here we have an optimisation solution buried in the ``pre_step`` events of the
    axis. We can get them all by simply calling:

    >>> sol['pre_step', 'series', 'minimize']
    <finesse.solutions.base.SolutionSet object at ...>

    You can see which solutions you have selected using:

    >>> sol['pre_step', 'series', 'minimize'].solutions
    [<OptimizeSolution of series/xaxis/pre_step/series/minimize @ ... children=0>,
     <OptimizeSolution of series/xaxis/pre_step/series/minimize @ ... children=0>,
     <OptimizeSolution of series/xaxis/pre_step/series/minimize @ ... children=0>]

    We can select a common attribute from these similar solutions by just acting
    on the SolutionSet, the attribute request is evaluated on each Solution present in
    the set and returned. For example, we can get the `result` attribute from each
    `OptimisationSolution` using:

    >>> print(sol['pre_step', 'series', 'minimize'].result)
    [ final_simplex: (array([[0.00e+00],
            [6.25e-05]]), array([0.00e+00, 6.25e-05]))
                fun: 0.0
            message: 'Optimization terminated successfully.'
               nfev: 6
                nit: 3
             status: 0
            success: True
                  x: array([0.])
      ...
    ]

    The returned attribute request will be a numpy array of objects or numerical
    values. This means it is possible to easily extract an array of values from
    a set of nested solutions.

    Which returns a tuple of the requested attributes from each of the nested solutions.
    You can also slice the solution object, which will again returns a reduced
    SolutionSet:

    >>> sol['pre_step', 'series', 'minimize'][::2]
    <finesse.solutions.base.SolutionSet object at ...>

    Each individual solution can be extracted using the `SolutionSet.solutions`
    attribute which returns a list of solution you can iterate over.
    """

    def __init__(self, solutions):
        if not isinstance(solutions, Iterable):
            raise Exception(f"{solutions} is not iterable")

        self.solutions = lst = []
        for value in solutions:
            if value not in lst:
                lst.append(value)

    def __iter__(self):
        return iter(self.solutions)

    def __contains__(self, value):
        return value in self.solutions

    def __len__(self):
        return len(self.solutions)

    def __getattr__(self, key):
        return np.asarray(tuple(getattr(_, key) for _ in self.solutions))

    def __getitem__(self, key):
        if isinstance(key, slice):
            return SolutionSet(self.solutions[key])
        else:
            return SolutionSet(_[key] for _ in self.solutions)


cdef class BaseSolution(ParameterChangingTreeNode):
    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.time = 0

    def __str__(self):
        def fn_name(child):
            if type(child) is not BaseSolution:
                r = f" - {child.__class__.__name__}"
            else:
                r = ""
            return child.name + r

        return self.draw_tree(fn_name, title="Solution Tree")

    def __getitem__(self, key):
        try:
            if isinstance(key, (slice, int)):
                return self.children[key]
            elif isinstance(key, str) or not isinstance(key, Iterable):
                rtn = tuple(child for child in self.children if child.name == str(key))
                if len(rtn) == 1:
                    return rtn[0]
                else:
                    # return specific set for certain solutions
                    if isinstance(rtn[0], ArraySolution):
                        return ArraySolutionSet(rtn)
                    else:
                        return SolutionSet(rtn)
            else:  # We're an iterable that isn't a string
                if len(key) == 1:
                    return self[key[0]]
                elif isinstance(key[0], slice):
                    raise NotImplementedError(
                        "Solution indexing only supports slices at the end of a key"
                    )
                else:
                    return self[key[0]][key[1:]]
        except IndexError as ex:
            # One last try if its us
            if isinstance(key, str) and self.name == key:
                return self
            raise FinesseException(f"Could not find an output `{key}` in the solution object {repr(self)}")

    def __repr__(self):
        return f"<{self.__class__.__name__} of {self.get_path()} @ {hex(id(self))} children={len(self.children)}>"

    # NOTE [ssl]: this exists to allow the CLI to plot any solution (so it doesn't need
    # to know ahead of time if a given action is plottable). In general plotting across
    # the solution classes should be made more consistent (see #191).
    def plot(self, *args, show=True, **kwargs):
        """Plot solution(s).

        If the solution contains child solutions, they are plotted in order. Solutions
        without plot arguments are skipped. Positional arguments passed to this method
        are assumed to be :class:`dict` and are unpacked into calls to each child's plot
        method. Global arguments to be passed to all plot methods can be specified
        directly as keyword arguments. Duplicate arguments specified in a positional
        argument :class:`dictionaries <dict>` take precendence over global arguments.

        Parameters
        ----------
        show : :class:`bool`, optional
            Show the figures.

        Other Parameters
        ----------------
        \\*args
            Sequence of :class:`dict` to use as parameters for the call to each child
            solution's `plot` method.

        \\**kwargs
            Keyword arguments supported by the child solution(s).

        Notes
        -----
        If the Nth solution contains no plot method, it still consumes the Nth
        positional argument passed to this method.
        """
        import matplotlib.pyplot as plt

        for child, child_kwargs in zip_longest(self.children, args, fillvalue={}):
            if not hasattr(child, "plot"):
                LOGGER.info(f"Skipping {child} as it is not plottable")
                continue

            # Merge child kwargs with global kwargs, overwriting duplicates in favour
            # of the former.
            child_kwargs = {**kwargs, **child_kwargs}

            if "show" not in child_kwargs:
                # Avoid blocking execution, only showing at the end, unless explicitly
                # set in the child kwargs.
                child_kwargs["show"] = False

            child.plot(**child_kwargs)

        if show:
            plt.show()



from contextlib import nullcontext

import numpy as np
import pytest

import finesse
from finesse.components import Laser, Modulator, SignalGenerator
from finesse.analysis.actions import DCFields
from finesse.enums import ModulatorType


@pytest.mark.parametrize(
    "mod_type, signal_node, raises",
    (
        (ModulatorType.am, "eom.amp.i", pytest.raises(NotImplementedError)),
        (ModulatorType.am, "eom.phs.i", pytest.raises(NotImplementedError)),
        (ModulatorType.pm, "eom.amp.i", nullcontext()),
        (ModulatorType.pm, "eom.phs.i", pytest.raises(NotImplementedError)),
    ),
)
def test_not_implemented_signal(mod_type, signal_node, raises):
    model = finesse.Model()
    model.add(Laser("l1"))
    model.add(Modulator("eom", f=100, midx=1, order=1, mod_type=mod_type))
    model.add(SignalGenerator("sgen", node=model.get(signal_node)))
    model.fsig.f = 1

    with raises:
        model.run()


def test_modulator_positive_only(model):
    """Test that the positive only option generates only the positive sidebands"""
    model.add(Laser("Laser", P=1))
    model.add(Modulator("Mod", 9e6, 0.2, order=2))
    model.connect(model.Laser.p1, model.Mod.p1)
    sol1 = model.run(DCFields())
    model.Mod.positive_only = True
    sol2 = model.run(DCFields())
    node = model.Mod.p2.o
    pos = sol1[node].squeeze()
    pos[sol1.frequencies < 0] = 0
    pos[0] = 1 - 0.5 * (1 - pos[0])
    np.testing.assert_allclose(pos, sol2[node].squeeze())



import pytest
from finesse.script.exceptions import KatParsingError
from finesse.script.compiler import KatIncorrectFormError
from testutils.text import dedent_multiline, escape_full


# Override default element to take two model parameter arguments.
@pytest.fixture
def fake_element_cls(fake_element_cls, fake_float_parameter):
    @fake_float_parameter("a", "Fake Parameter A")
    @fake_float_parameter("b", "Fake Parameter B")
    class FakeElement(fake_element_cls):
        def __init__(self, name, a=None, b=None):
            super().__init__(name)
            self.a = a
            self.b = b

    return FakeElement


@pytest.fixture
def spec(
    spec,
    set_spec_constructs,
    fake_element_adapter_factory,
    fake_element_cls,
    finesse_binop_sub,
):
    spec.register_element(
        fake_element_adapter_factory(
            fake_element_cls,
            other_names=["fake_element_alias"],  # necessary for 'test_aliasing'
        )
    )
    # Have to use real Finesse operator here because the builder matches against Finesse
    # operations.
    set_spec_constructs("binary_operators", {"-": finesse_binop_sub})

    return spec


@pytest.mark.parametrize(
    "script,element_def",
    (
        pytest.param(
            "fake_element myelement 1",
            ("myelement", 1),
            id="element-with-arg",
        ),
        pytest.param(
            "fake_element myelement a=1",
            ("myelement", 1),
            id="element-with-kwarg",
        ),
    ),
)
def test_element(model_matcher, fake_element_cls, script, element_def):
    model_matcher(script, [fake_element_cls(*element_def)])


@pytest.mark.parametrize(
    "script,element_defs",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 1
                fake_element myelement2 1-myelement1.a
                """
            ),
            [("myelement1", 1), ("myelement2", 0)],
        ),
    ),
)
def test_argument_reference(model_matcher, fake_element_cls, script, element_defs):
    model_matcher(script, [fake_element_cls(*defs) for defs in element_defs])


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "fake_element myelement1 myelement1.a",
            (
                "\nline 1: cannot set myelement1.a to self-referencing value myelement1.a\n"
                "-->1: fake_element myelement1 myelement1.a\n"
                "                              ^^^^^^^^^^^^"
            ),
            id="arg-self-ref",
        ),
        pytest.param(
            "fake_element myelement1 1-myelement1.a",
            (
                "\nline 1: cannot set myelement1.a to self-referencing value (1-myelement1.a)\n"
                "-->1: fake_element myelement1 1-myelement1.a\n"
                "                                ^^^^^^^^^^^^"
            ),
            id="arg-expr-self-ref",
        ),
        pytest.param(
            "fake_element myelement1 a=myelement1.a",
            (
                "\nline 1: cannot set myelement1.a to self-referencing value myelement1.a\n"
                "-->1: fake_element myelement1 a=myelement1.a\n"
                "                                ^^^^^^^^^^^^"
            ),
            id="kwarg-self-ref",
        ),
        pytest.param(
            "fake_element myelement1 a=1-myelement1.a",
            (
                "\nline 1: cannot set myelement1.a to self-referencing value (1-myelement1.a)\n"
                "-->1: fake_element myelement1 a=1-myelement1.a\n"
                "                                  ^^^^^^^^^^^^"
            ),
            id="kwarg-expr-self-ref",
        ),
    ),
)
def test_directly_self_referencing_parameter_invalid(compiler, script, error):
    with pytest.raises(KatParsingError, match=escape_full(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script, error",
    (
        pytest.param(
            "fake_element",
            (
                "\nline 1: 'fake_element' should be written in the form 'fake_element name a=none b=none'\n"
                "-->1: fake_element\n"
                "      ^^^^^^^^^^^^"
            ),
            id="kwarg-expr-self-ref",
        ),
    ),
)
def test_incorrect_form_error(compiler, script, error):
    with pytest.raises(KatIncorrectFormError, match=escape_full(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,element_values",
    (
        pytest.param(
            "fake_element myelement1 1 0.5-myelement1.a",
            {"a": 1, "b": -0.5},
            id="arg-self-ref-to-arg",
        ),
        pytest.param(
            "fake_element myelement1 1 b=0.5-myelement1.a",
            {"a": 1, "b": -0.5},
            id="kwarg-self-ref-to-arg",
        ),
        pytest.param(
            "fake_element myelement1 a=1 b=0.5-myelement1.a",
            {"a": 1, "b": -0.5},
            id="kwarg-self-ref-to-kwarg",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 a=myelement2.a b=0.5-myelement1.a
                fake_element myelement2 a=0.5-myelement2.b b=1
                """
            ),
            {"a": -0.5, "b": 1},
            id="second-order-self-ref",
        ),
    ),
)
def test_same_element_referencing_element(compiler, script, element_values):
    model = compiler.compile(script)
    for key, value in element_values.items():
        assert model.get(f"myelement1.{key}").eval() == value


@pytest.mark.parametrize(
    "script,element_defs",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 1
                fake_element myelement2 2
                fake_element myelement3 3
                """
            ),
            [
                ("myelement1", 1),
                ("myelement2", 2),
                ("myelement3", 3),
            ],
            id="elements",
        ),
    ),
)
def test_script(model_matcher, fake_element_cls, script, element_defs):
    model_matcher(script, [fake_element_cls(*defs) for defs in element_defs])


@pytest.mark.parametrize(
    "script",
    (
        "fake_element_alias myelement 1",
        "fake_element myelement 1",
    ),
)
def test_aliasing(script, model_matcher, fake_element_cls):
    elements = [fake_element_cls("myelement", 1)]
    model_matcher(script, elements)



import numpy as np
import finesse
from finesse.utilities.maps import circular_aperture, surface_point_absorber
from finesse.knm.maps import (
    Map,
    map_scattering_coefficients,
    scattering_coefficients_to_KnmMatrix,
)
from finesse.knm.tools import make_bayerhelms_matrix
import pytest


def test_custom_map():
    """Really just testing that these class methods get called without error."""

    class CustomMap(Map):
        def __init__(self):
            x = np.linspace(-5e-2, 5e-2, num=200)
            y = np.linspace(-5e-2, 5e-2, num=201)

            super().__init__(
                x,
                y,
                opd=self.surface,
                amplitude=self.aperture,
            )

        def aperture(self, model=None):
            return np.ones_like(self.R)

        def surface(self, model=None):
            return np.zeros_like(self.R)

    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    m m1 L=40e-6 T=0.014 Rc=-1945
    """
    )

    model.m1.surface_map = CustomMap()
    model.run()


def test_changing_map():
    """Tests that changing maps just run, not comparing any outputs at the moment."""
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    m m1 L=40e-6 T=0.014 Rc=-1945
    m m2 L=40e-6 T=0 Rc=2245
    link(l1, m1, 3994, m2)
    cav c m2.p1.o

    pd P m1.p2.i

    modes(maxtem=7)

    var offset 0 # point absorber position
    """
    )

    R_ap = 0.4 / 2  # test mass radius
    x = y = np.linspace(-R_ap, R_ap, 400)

    def pabs_map(_map, model):
        h = 200e-3
        w = 20e-6
        power_absorbed = 20e-3

        return surface_point_absorber(
            _map.x - float(model.offset), _map.y, w, h, power_absorbed
        )

    itm_map = Map(x, y, amplitude=circular_aperture(x, y, R_ap))
    etm_map = Map(x, y, amplitude=itm_map.amplitude, opd=pabs_map)
    # changing maps aren't picked up by refill checks
    # need to force m2 to refill at the moment
    model.m2.phi.is_tunable = True
    model.m2.surface_map = etm_map
    model.run("xaxis(offset, lin, 0, 0.08, 2, pre_step=update_maps())")


def test_remove_tilts():
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    m m1 L=0e-6 T=0 Rc=1945
    link(l1, m1)
    gauss g1 l1.p1.o w0=0.012044800081073403 z=-1833.9474927802573

    ad a m1.p1.o n=0 m=1 f=0
    ad b m1.p1.o n=1 m=0 f=0

    modes(maxtem=2)
    var tilt 0
    """
    )

    R_ap = 0.4 / 2  # test mass radius
    x = y = np.linspace(-R_ap, R_ap, 100)

    def tilt_map(_map, model):
        return _map.X * float(model.tilt) + _map.Y * float(-model.tilt * 0.033)

    # changing maps aren't picked up by refill checks
    # need to force m2 to refill at the moment
    model.m1.phi.is_tunable = True
    model.m1.surface_map = Map(x, y, opd=tilt_map, auto_remove_tilts=True)

    out = model.run("xaxis(tilt, log, 1e-12, 1e-6, 10, pre_step=update_maps())")
    # roughly limit to make sure all the tilt is removed
    assert np.all(abs(out["a"]) < 1e-13)
    assert np.all(abs(out["b"]) < 1e-13)


def test_remove_astigmatism():
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    m m1 L=0e-6 T=0
    link(l1, m1)
    gauss g1 l1.p1.o w0=0.05 z=0

    ad a m1.p1.o n=0 m=2 f=0
    ad b m1.p1.o n=2 m=0 f=0

    modes(maxtem=2)
    var tilt 0
    """
    )

    R_ap = 0.4 / 2  # test mass radius
    x = y = np.linspace(-R_ap, R_ap, 100)

    def tilt_map(_map, model):
        return _map.X**2 * float(model.tilt) + _map.Y**2 * float(-model.tilt * 0.033)

    # changing maps aren't picked up by refill checks
    # need to force m2 to refill at the moment
    model.m1.phi.is_tunable = True
    model.m1.surface_map = Map(x, y, opd=tilt_map, auto_remove_astigmatism=True)

    out = model.run("xaxis(tilt, log, 1e-12, 1e-6, 10, pre_step=update_maps())")
    # roughly limit to make sure all the tilt is removed
    assert np.all(abs(out["a"]) < 1e-13)
    assert np.all(abs(out["b"]) < 1e-13)


@pytest.mark.parametrize("w", [None, 0.01, 0.05])
def test_remove_piston(w):
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    m m1 L=0 T=0
    link(l1, m1)
    gauss g1 l1.p1.o w0=0.05 z=0

    ad a m1.p1.o n=0 m=2 f=0
    ad b m1.p1.o n=2 m=0 f=0

    modes(maxtem=2)
    var tilt 0
    """
    )

    R_ap = 0.4 / 2  # test mass radius
    x = y = np.linspace(-R_ap, R_ap, 500)

    model.m1.surface_map = Map(x, y, opd=10 * np.ones((500, 500)))
    model.m1.surface_map.remove_piston(w)
    assert np.all(model.m1.surface_map.opd < 1e-6)


def test_remove_curvatures():
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    m m1 L=0e-6 T=0
    link(l1, m1)
    gauss g1 l1.p1.o w0=0.05 z=0

    ad a m1.p1.o n=0 m=2 f=0
    ad b m1.p1.o n=2 m=0 f=0

    modes(maxtem=2)
    var tilt 0
    """
    )

    R_ap = 0.4 / 2  # test mass radius
    x = y = np.linspace(-R_ap, R_ap, 100)

    def tilt_map(_map, model):
        return _map.X**2 * float(model.tilt) + _map.Y**2 * float(model.tilt)

    # changing maps aren't picked up by refill checks
    # need to force m2 to refill at the moment
    model.m1.phi.is_tunable = True
    model.m1.surface_map = Map(x, y, opd=tilt_map, auto_remove_curvatures=True)

    out = model.run("xaxis(tilt, log, 1e-12, 1e-6, 10, pre_step=update_maps())")
    # roughly limit to make sure all the tilt is removed
    assert np.all(abs(out["a"]) < 1e-13)
    assert np.all(abs(out["b"]) < 1e-13)


def test_reverse_gouy():
    """Curved beam reflected from a flat tilted mirror, check that BH and maps calculate
    same output with reverse gouy."""
    model = finesse.Model()
    model.parse(
        """
    l l1
    m m1 R=1 T=0
    link(l1, m1)
    fd Er m1.p1.o 0

    modes(maxtem=1)
    gauss g1 m1.p1.i w=10e-3 Rc=3000
    """
    )
    model.beam_trace()
    qx1, qy1 = model.m1.p1.i.q
    qx2, qy2 = -np.conj(model.m1.p1.i.q)
    Kbh = make_bayerhelms_matrix(
        qx1, qx2, qy1, qy2, 2e-6, 0, select=model.homs, reverse_gouy=True
    )

    x = np.linspace(-0.04, 0.04, 201)
    y = np.linspace(-0.04, 0.04, 201)
    X, Y = np.meshgrid(x, y)
    Z = -1e-6 * X
    smap = Map(x, y, opd=Z)

    K = map_scattering_coefficients(
        (qx1, qy1, qx2, qy2),
        model.modes_setting["maxtem"],
        smap.x.copy(),
        smap.y.copy(),
        smap.get_z(2 * np.pi / 1064e-9, -2, model),
        reverse_gouy=True,
    )

    K_map = scattering_coefficients_to_KnmMatrix(model.homs, K)

    assert abs(Kbh.data - K_map.data).max() < 1e-13


def test_flip_lr():
    model = finesse.Model()
    model.parse(
        """
    l l1
    m m1 R=1 T=0
    link(l1, m1)
    fd Er m1.p1.o 0

    modes(maxtem=1)
    gauss g1 m1.p1.i w=10e-3 Rc=3000
    """
    )
    model.beam_trace()
    qx1, qy1 = model.m1.p1.i.q
    qx2, qy2 = -np.conj(model.m1.p1.i.q)

    x = np.linspace(-0.04, 0.04, 201)
    y = np.linspace(-0.04, 0.04, 201)
    X, Y = np.meshgrid(x, y)
    Z = 1e-9 * X
    smap = Map(x, y, opd=Z)

    KA = map_scattering_coefficients(
        (qx1, qy1, qx2, qy2),
        model.modes_setting["maxtem"],
        smap.x.copy(),
        smap.y.copy(),
        smap.get_z(2 * np.pi / 1064e-9, -2, model),
        reverse_gouy=True,
        flip_lr=False,
    )

    KB = map_scattering_coefficients(
        (qx1, qy1, qx2, qy2),
        model.modes_setting["maxtem"],
        smap.x.copy(),
        smap.y.copy(),
        smap.get_z(2 * np.pi / 1064e-9, -2, model),
        reverse_gouy=True,
        flip_lr=True,
    )

    K_mapA = scattering_coefficients_to_KnmMatrix(model.homs, KA)
    K_mapB = scattering_coefficients_to_KnmMatrix(model.homs, KB)

    assert np.allclose(K_mapA.data[1, 0], -K_mapB.data[1, 0], atol=1e-13)
    assert np.allclose(K_mapA.data[0, 1], -K_mapB.data[0, 1], atol=1e-13)


def test_curved_map():
    x = np.linspace(-5e-3, 5e-3, num=200)
    y = np.linspace(-5e-3, 5e-3, num=200)
    xv, yv = np.meshgrid(x, y)
    R2 = xv**2 + yv**2
    Rc = 1

    model = finesse.Model()
    model.parse(
        """
    l l1
    l l2
    m m1 R=1 T=0 Rc=1
    link(l1, m1, l2)
    pd P1 m1.p1.o
    pd P2 m1.p2.o
    """
    )
    model.l1.p1.o.q = finesse.BeamParam(w=1e-3, Rc=1)
    model.l2.p1.o.q = finesse.BeamParam(w=1e-3, Rc=-1)
    model.modes(maxtem=2)
    model.m1.surface_map = Map(x, y, opd=R2 / (2 * Rc), is_focusing_element=True)
    assert abs(1 - model.run()["P1"]) < 1e-14
    assert abs(1 - model.run()["P2"]) < 1e-14


def test_astigmatic_lens():
    x = np.linspace(-5e-3, 5e-3, num=200)
    y = np.linspace(-5e-3, 5e-3, num=200)
    xv, yv = np.meshgrid(x, y)
    X2 = xv**2
    Y2 = yv**2

    model = finesse.Model()
    model.parse(
        """
    l l1
    l l2
    alens L1 fx=100 fy=200
    link(l1, L1, l2)
    pd P1 L1.p1.o
    pd P2 L1.p2.o
    """
    )
    model.l1.p1.o.q = finesse.BeamParam(w=1e-3, Rc=1)
    model.modes(maxtem=2)
    model.L1.OPD_map = finesse.knm.Map(
        x,
        y,
        opd=-X2 / (2 * 100) - Y2 / (2 * 200),
        is_focusing_element=True,
    )
    assert abs(1 - model.run()["P1"]) < 1e-14
    assert abs(1 - model.run()["P2"]) < 1e-14


def test_thin_lens():
    x = np.linspace(-5e-3, 5e-3, num=200)
    y = np.linspace(-5e-3, 5e-3, num=200)
    xv, yv = np.meshgrid(x, y)
    R2 = xv**2 + yv**2

    model = finesse.Model()
    model.parse(
        """
    l l1
    l l2
    lens L1 f=100
    link(l1, L1, l2)
    pd P1 L1.p1.o
    pd P2 L1.p2.o
    """
    )
    model.l1.p1.o.q = finesse.BeamParam(w=1e-3, Rc=1)
    model.modes(maxtem=2)
    model.L1.OPD_map = finesse.knm.Map(
        x, y, opd=-R2 / (2 * 100), is_focusing_element=True
    )
    assert abs(1 - model.run()["P1"]) < 1e-14
    assert abs(1 - model.run()["P2"]) < 1e-14


def test_lens_remove_curvature_put_into_lens():
    """Put map focal length into a lens object then ensure that mode-matching is
    calculated properly."""
    x = np.linspace(-5e-3, 5e-3, num=200)
    y = np.linspace(-5e-3, 5e-3, num=200)
    xv, yv = np.meshgrid(x, y)
    R2 = xv**2 + yv**2

    model = finesse.Model()
    model.parse(
        """
    l l1
    l l2
    lens L1 f=100
    link(l1, L1, l2)
    pd P1 L1.p1.o
    pd P2 L1.p2.o
    """
    )
    model.l1.p1.o.q = finesse.BeamParam(w=1e-3, Rc=1)
    model.modes(maxtem=2)
    model.L1.OPD_map = finesse.knm.Map(
        x, y, opd=-R2 / (2 * 200), put_focal_length=model.L1
    )
    model.L1.OPD_map.remove_curvatures(1e-3)
    assert abs(1 - model.run()["P1"]) < 1e-14
    assert abs(1 - model.run()["P2"]) < 1e-14



from finesse.solutions.base import BaseSolution
from finesse.solutions.base cimport BaseSolution
from cpython.ref cimport PyObject

import numpy as np
cimport numpy as np


cdef class ArraySolution(BaseSolution):
    cdef:
        readonly dict trace_info
        readonly dict axis_info
        readonly np.ndarray _data
        readonly np.dtype dtype
        readonly int _axes
        readonly int _num
        readonly tuple shape
        readonly tuple x
        readonly tuple params
        readonly tuple detectors
        readonly bint masked
        PyObject** workspaces
        Py_ssize_t num_workspaces
        bint enabled
        readonly np.ndarray flatiters

    cpdef int update(self, int index, bint mask) except -1
    cpdef enable_update(self, detector_workspaces)



import numpy as np
import pytest

import finesse
from finesse.analysis.actions import Maximize, Minimize
from finesse.analysis.actions.optimisation import OptimizationWarning


def test_minimize_basic():
    # Test basic minimisation that uses nelder-mead arguments
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    pd P l1.p1.o
    minimize(P, l1.P)
    """
    )
    sol = model.run()
    assert sol.x < 1e-2


def test_minimize_basic_nelder_mead():
    # Test basic minimisation that uses nelder-mead arguments
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    pd P l1.p1.o
    minimize(P, l1.P, xatol=1e-6, adaptive=True)
    """
    )
    sol = model.run()
    assert sol.x < 1e-6


# https://gitlab.com/ifosim/finesse/finesse3/-/issues/810
@pytest.mark.parametrize("method", (None, "nelder-mead"))
def test_minimize_tol(method):
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    pd P l1.p1.o
    """
    )
    sol = model.run(
        Minimize(model.P, model.l1.P, method=method, bounds=[-2, 2], tol=1e-8)
    )
    assert sol.x < 1e-6


@pytest.fixture
def converge_first_value():
    model = finesse.Model()
    model.parse(
        """
        laser l1 P=1
        link(l1, ITM)
        m ITM
        link(ITM, ETM)
        m ETM
        pd circ ITM.p2.o
    """
    )
    action = Maximize(model.circ, model.ETM.phi, verbose=True)
    return model, action


# https://gitlab.com/ifosim/finesse/finesse3/-/issues/635
def test_maximize_converge_first_value(converge_first_value):
    model, action = converge_first_value
    assert model.ETM.phi.value == 0
    sol = model.run(action)
    assert sol.x[0] == 0
    assert model.ETM.phi.value == 0


def test_single_iteration_warning(converge_first_value):
    model, action = converge_first_value
    model.ETM.phi.value = 1e-3
    with pytest.warns(OptimizationWarning):
        model.run(action)


def test_minimize_RF_readout_phase():
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    mod mod1 f=10M midx=0.1 mod_type=am
    link(l1, mod1)
    pd1 RF mod1.p2.o 10M 0
    minimize(RF, RF.phase, xatol=1e-6, adaptive=True)
    """
    )
    sol = model.run()
    # Max solution should be at 90 phase
    assert np.allclose(sol.x, 90, atol=1e-6)


def test_maximize_RF_readout_phase():
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    mod mod1 f=10M midx=0.1 mod_type=am
    link(l1, mod1)
    pd1 RF mod1.p2.o 10M 0
    maximize(RF, RF.phase, xatol=1e-6, adaptive=True)
    """
    )
    sol = model.run()
    # Max solution should be at 0 phase
    assert np.allclose(sol.x, 0, atol=1e-6)


def test_maximize_multiple_targets():
    """Coupled cavity optimize."""
    model = finesse.Model()
    model.parse(
        """
    l l1 P=1
    m m1 R=0.98 T=0.02 phi=10
    m m2 R=0.99 T=0.01
    m m3 R=1 T=0 phi=-20
    link(l1, m1, m2, m3)
    pd P m3.p1.i

    maximize(P, [m1.phi, m3.phi], xatol=1e-7, adaptive=True)
    """
    )
    sol = model.run()
    assert np.allclose(sol.x, [90, 0], atol=1e-6)


def test_bounds():
    model = finesse.Model()
    model.parse(
        """
    l l1
    l l2
    bs BS R=0.5 T=0.5
    link(l1, BS.p1)
    link(l2, BS.p4)
    pd P BS.p2.o
    """
    )

    sol = model.run(
        """
    series(
        maximize(P, [l1.P, l2.P, l1.phase], bounds=[[0,4], [0, 4], [-180, 180]]),
        noxaxis()
    )
    """
    )

    assert abs(sol["noxaxis"]["P"] - 8) < 1e-6


def test_optimise_squeezing_modematch():
    kat = finesse.script.parse(
        """
        l l1 P=1e14
        s s1 l1.p1 bs1.p4

        #squeezed source
        sq sq1 db=20
        s s2 sq1.p1 dbs1.p1

        #first filter cavity
        dbs dbs1
        s s21 dbs1.p3 ITM.p1
        m ITM R=0.99 T=0.01 Rc=-2.5
        s sCAV ITM.p2 ETM.p1 L=1
        m ETM R=1 T=0.0 Rc=2.5

        #----------------------------------------
        # Mode matching telescope
        #----------------------------------------
        s st1 dbs1.p4 lens1.p1 L=1
        lens lens1 f=1
        s st2 lens1.p2 lens2.p1 L=1
        lens lens2 f=1
        s st3 lens2.p2 dbsf.p1 L=1
        #----------------------------------------

        #second filter cavity
        dbs dbsf
        s s21f dbsf.p3 ITMf.p1
        m ITMf R=0.99 L=0 Rc=-2.5
        s sCAVf ITMf.p2 ETMf.p1 L=1
        m ETMf R=1 T=0.0 Rc=2.5
        s s4f dbsf.p4 bs1.p1

        #homodyne
        bs bs1 T=1e-14 L=0 alpha=45 phi=-4.5

        #cavities
        cav FP ITM.p2.o priority=0
        cav FP2 ITMf.p2.o priority=1

        fsig(1)

        qnoised sqzd_noise bs1.p2.o

        modes(maxtem=4)
        """
    )
    sol0 = kat.run()
    opt = kat.run("minimize(sqzd_noise, [lens1.f, lens2.f])")
    kat.run("minimize(sqzd_noise, l1.phase)")
    sol1 = kat.run()
    assert opt.x[0] != 1
    assert opt.x[1] != 1
    assert sol1["sqzd_noise"] < sol0["sqzd_noise"]
    assert kat.cavity_mismatch("FP", "FP2")[0] < 1e-5
    assert kat.cavity_mismatch("FP", "FP2")[1] < 1e-5



from cpython.ref cimport PyObject
from finesse.simulations.simulation cimport BaseSimulation

from .cyexpr cimport (
    cy_expr,
    cy_expr_new,
    cy_expr_init,
    cy_expr_free,
    cy_expr_eval,
)

cdef class BaseCValues:
    cdef:
        Py_ssize_t N
        double **ptr
        tuple params

    cdef setup(self, tuple params, Py_ssize_t data_size, double** data_start)
    cdef get_param_ptr(self, unicode name, double**ptr)


cdef class ElementWorkspace:
    cdef:
        readonly BaseSimulation sim # simulation object this workspace is attached to
        readonly bint _unique_element # Only a single element of this type can be added to a model
        readonly object owner # Model element that owns this workspace
        public int owner_id # Owner's id for this particular simualation
        readonly object values
        bint is_c_values
        type type_c_values
        int num_changing_parameters
        PyObject** chprm
        double** chprm_target
        cy_expr** chprm_expr # Compiled changing symbolic expressions array
        bint first

    cpdef int compile_cy_exprs(self) except -1
    cpdef update_parameter_values(self)



import pytest
import finesse
import numpy as np
from finesse.knm.bayerhelms import make_bayerhelms_matrix


def test_mirror_pitch_yaw_mode_content():
    """This reflects a perfectly aligned beam from a mirror.

    The beam has a waist at the mirror. The mirror is misaligned by a positive
    pitch/yaw. As we have right-handed coordinates for the components, this means the
    beam is deflected in a negative direction in x/y. Or this means that the 10/01 mode
    content has a negative sign.
    """
    IFO = finesse.Model()

    IFO.parse(
        """
        l l1 P=1
        s s1 l1.p1 m1.p1 L=1e15
        m m1 R=1 T=0 xbeta=1e-9 ybeta=1e-9
        s s2 l2.p1 m1.p2 L=1e15
        l l2 P=1

        gauss g1 m1.p1.o z=0 w0=1e-3

        ad u100 l1.p1.i 0 n=0 m=0
        ad u200 l2.p1.i 0 n=0 m=0
        ad u110 l1.p1.i 0 n=1 m=0
        ad u210 l2.p1.i 0 n=1 m=0
        ad u101 l1.p1.i 0 n=0 m=1
        ad u201 l2.p1.i 0 n=0 m=1

        modes(maxtem=1)
        """
    )

    out = IFO.run()

    q = finesse.BeamParam(w0=1e-3, z=0)
    # factor of 2 because angle change is twice the misalignment
    a10 = 2 * IFO.m1.xbeta / q.divergence
    a01 = 2 * IFO.m1.ybeta / q.divergence

    # 00 power should be 1 - the power in ecah 10 01 mode
    assert abs(abs(out["u100"]) ** 2 - (1 - a10**2 - a01**2)) < 1e-15
    assert abs(abs(out["u200"]) ** 2 - (1 - a10**2 - a01**2)) < 1e-15

    # right-handed rotation pitch yaw tilts beam in negative directions
    # after propagation 1e15m it should all be real valued, so a pure displacement
    assert out["u110"] - (-a10) < 1e-15
    assert out["u101"] - (-a01) < 1e-15
    # Yaw other side is same negative sign, but pitch is reversed
    # and beam goes up
    assert out["u210"] - (-a10) < 1e-15
    assert out["u201"] - (+a01) < 1e-15


def test_mirror_z_pitch_yaw_forces_side1():
    IFO = finesse.Model()

    IFO.parse(
        f"""
        l l1 P=1
        s s1 l1.p1 m1.p1 L=0
        m m1 R=1 T=0
        s s2 l2.p1 m1.p2 L=0
        l l2 P=1

        gauss g1 m1.p1.o z=0 w0=1e-3
        free_mass m1_sus m1.mech mass=inf I_yaw=inf I_pitch=inf

        ad u100 m1.p1.i 0 n=0 m=0
        ad u200 m1.p2.i 0 n=0 m=0
        ad u110 m1.p1.i 0 n=1 m=0
        ad u210 m1.p2.i 0 n=1 m=0
        ad u101 m1.p1.i 0 n=0 m=1
        ad u201 m1.p2.i 0 n=0 m=1

        ad Fz m1.mech.F_z fsig.f
        ad Fyaw m1.mech.F_yaw fsig.f
        ad Fpitch m1.mech.F_pitch fsig.f

        sgen signal l1.pwr

        fsig({1/(2*np.pi)})
        modes(maxtem=1)
    """
    )
    IFO.l1.tem(1, 0, 1e-3, 0)
    IFO.l1.tem(0, 1, 1e-3, 0)

    q = finesse.BeamParam(w0=1e-3, z=0)

    out = IFO.run()

    dx = out["u110"] * q.w0
    # dy = out["u101"] * q.w0

    # negative here because mirror z direction is opposite to beams
    # so an increase in power applies a neative force
    Fz = -2 * IFO.l1.P / 299792458
    # positive beam x displacements should generate
    # a positive torque on the mirror from right-handed
    # rotation
    Fyaw = dx * 2 * IFO.l1.P / 299792458
    # a positive beam displacement in y should be negative...
    Fpitch = -dx * 2 * IFO.l1.P / 299792458

    assert abs(out["Fz"] - Fz) < 1e-15
    assert abs(out["Fyaw"] - Fyaw) < 1e-15
    assert abs(out["Fpitch"] - Fpitch) < 1e-15


def test_mirror_z_pitch_yaw_forces_side2():
    IFO = finesse.Model()

    IFO.parse(
        f"""
        l l1 P=1
        s s1 l1.p1 m1.p1 L=0
        m m1 R=1 T=0
        s s2 l2.p1 m1.p2 L=0
        l l2 P=1

        gauss g1 m1.p1.o z=0 w0=1e-3
        free_mass m1_sus m1.mech mass=inf I_yaw=inf I_pitch=inf

        ad u100 m1.p1.i 0 n=0 m=0
        ad u200 m1.p2.i 0 n=0 m=0
        ad u110 m1.p1.i 0 n=1 m=0
        ad u210 m1.p2.i 0 n=1 m=0
        ad u101 m1.p1.i 0 n=0 m=1
        ad u201 m1.p2.i 0 n=0 m=1

        ad Fz m1.mech.F_z fsig.f
        ad Fyaw m1.mech.F_yaw fsig.f
        ad Fpitch m1.mech.F_pitch fsig.f

        sgen signal l2.pwr

        fsig({1/(2*np.pi)})
        modes(maxtem=1)
        """
    )
    IFO.l1.tem(1, 0, 1e-3, 0)
    IFO.l1.tem(0, 1, 1e-3, 0)

    q = finesse.BeamParam(w0=1e-3, z=0)

    out = IFO.run()

    dx = out["u210"] * q.w0
    # dy = out["u201"] * q.w0

    Fz = 2 * IFO.l1.P / 299792458
    # positive beam x displacements should generate
    # a positive torque on the mirror from right-handed
    # rotation
    Fyaw = dx * 2 * IFO.l1.P / 299792458
    Fpitch = dx * 2 * IFO.l1.P / 299792458

    assert abs(out["Fz"] - Fz) < 1e-15
    assert abs(out["Fyaw"] - Fyaw) < 1e-15
    assert abs(out["Fpitch"] - Fpitch) < 1e-15


def test_mirror_xybeta_curved_mirror():
    q = finesse.BeamParam(w0=1e-3, z=100)

    IFO = finesse.Model()
    IFO.parse(
        f"""
        l l1 P=1
        s s1 l1.p1 m1.p1 L={q.z}
        m m1 R=1 T=0 Rc={q.Rc} xbeta=1e-9 ybeta=1e-9

        gauss g1 l1.p1.o z=0 w0=1e-3

        ad u100 l1.p1.i n=0 m=0 f=0
        ad u110 l1.p1.i n=1 m=0 f=0
        ad u101 l1.p1.i n=0 m=1 f=0

        modes(maxtem=1)
        """
    )

    q = finesse.BeamParam(w0=1e-3, z=0)

    out = IFO.run()

    dx = out["u110"].real * q.w0
    dy = out["u101"].real * q.w0

    # reflect beam from curved mirror some distance away.
    # back at the waist the beam should be displaced by the lever
    # arm and still tilted the same direction
    assert abs(dx - IFO.m1.xbeta * -2 * IFO.spaces.s1.L) < 1e-14
    assert abs(dy - IFO.m1.ybeta * -2 * IFO.spaces.s1.L) < 1e-14
    # accuracy on numerics here isn't as good
    assert abs(out["u101"].imag * q.divergence - IFO.m1.ybeta) < 1e-9
    assert abs(out["u110"].imag * q.divergence - IFO.m1.xbeta) < 1e-9


@pytest.fixture
def tilted_plate():
    """Transmissive 1m thick plate, refractive index = 2."""
    return """
    l l1 1 0 n0
    l l2 1 0 n5

    s s1 1 1 n0 n1
    m m1 0 1 0 n1 n2
    s s2 1 2 n2 n3
    m m2 0 1 0 n3 n4
    s s3 0 1 n4 n5

    maxtem 1
    """


def test_tilted_plate_offset_12_yaw(tilted_plate):
    # Tests beam gets displaced by a thick plate at an angle correctly
    # Going side 1 to 2
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    IFO.parse_legacy(
        f"""
        {tilted_plate}

        xaxis m1 xbeta lin -1n 1n 1
        put m2 xbeta $x1

        gauss g1 s3 n5 {q.w0} {q.z}
        ad a 1 0 0 n5
        ad c 0 0 0 n5

        yaxis abs:deg
        """
    )

    out = IFO.run()
    a = out["a"] * np.exp(-1j * np.angle(out["c"]))
    delta = (
        IFO.spaces.s2.L * IFO.analysis.stop * (1 - IFO.spaces.s1.nr / IFO.spaces.s2.nr)
    )
    assert all(abs(delta / q.w0 * np.array([1, -1]) - a) < 1e-13)


def test_tilted_plate_offset_21_yaw(tilted_plate):
    # Tests beam gets displaced by a thick plate at an angle correctly
    # Going side 2 to 1
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    IFO.parse_legacy(
        f"""
        {tilted_plate}

        xaxis m1 xbeta lin -1n 1n 1
        put m2 xbeta $x1

        gauss g1 s1 n0 {q.w0} {q.z}
        ad a 1 0 0 n0
        ad c 0 0 0 n0

        yaxis abs:deg
        """
    )

    out = IFO.run()
    a = out["a"] * np.exp(-1j * np.angle(out["c"]))
    delta = (
        IFO.spaces.s2.L * IFO.analysis.stop * (1 - IFO.spaces.s1.nr / IFO.spaces.s2.nr)
    )
    assert all(abs(delta / q.w0 * np.array([1, -1]) - a) < 1e-13)


def test_tilted_plate_offset_12_pitch(tilted_plate):
    # Tests beam gets displaced by a thick plate at an angle correctly
    # Going side 1 to 2
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    IFO.parse_legacy(
        f"""
        {tilted_plate}

        xaxis m1 ybeta lin -1n 1n 1
        put m2 ybeta $x1

        gauss g1 s3 n5 {q.w0} {q.z}
        ad a 0 1 0 n5
        ad c 0 0 0 n5

        yaxis abs:deg
        """
    )

    out = IFO.run()
    a = out["a"] * np.exp(-1j * np.angle(out["c"]))
    delta = (
        IFO.spaces.s2.L * IFO.analysis.stop * (1 - IFO.spaces.s1.nr / IFO.spaces.s2.nr)
    )
    assert all(abs(delta / q.w0 * np.array([-1, 1]) - a) < 1e-13)


def test_tilted_plate_offset_21_pitch(tilted_plate):
    # Tests beam gets displaced by a thick plate at an angle correctly
    # Going side 2 to 1
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    IFO.parse_legacy(
        f"""
        {tilted_plate}

        xaxis m1 ybeta lin -1n 1n 1
        put m2 ybeta $x1

        gauss g1 s1 n0 {q.w0} {q.z}
        ad a 0 1 0 n0
        ad c 0 0 0 n0

        yaxis abs:deg
        """
    )

    out = IFO.run()
    a = out["a"] * np.exp(-1j * np.angle(out["c"]))

    delta = (
        IFO.spaces.s2.L * IFO.analysis.stop * (1 - IFO.spaces.s1.nr / IFO.spaces.s2.nr)
    )
    assert all(abs(delta / q.w0 * np.array([1, -1]) - a) < 1e-13)


def test_tilted_beam_incident_on_normal_plate():
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    n1 = 1
    n2 = 2

    IFO.parse_legacy(
        f"""
        l l1 1 0 n0
        s s1 0 {n1} n0 n1
        m m1 0 1 0 n1 n2
        s s2 0.1 {n2} n2 n3
        m m2 0 1 0 n3 n4
        s s3 0 {n1} n4 n5
        l l2 1 0 n5

        maxtem 1
        noxaxis

        gauss g1 l1 n0 {q.w0} {q.z}

        ad a 0 1 0 n1*
        ad ca 0 0 0 n1*
        ad b 0 1 0 n2
        ad cb 0 0 0 n2
        ad c 0 1 0 n3*
        ad cc 0 0 0 n3*
        ad d 0 1 0 n4
        ad cd 0 0 0 n4
        """
    )

    # Perform a beam trace to allow us to grab the divergence below.
    IFO.beam_trace()
    IFO.l1.tem(0, 1, (4.2222222222e-13 / IFO.l1.p1.o.qx.divergence) ** 2, 90)
    out = IFO.run()

    a = out["a"] * np.exp(-1j * np.angle(out["ca"]))
    b = out["b"] * np.exp(-1j * np.angle(out["cb"]))
    c = out["c"] * np.exp(-1j * np.angle(out["cc"]))
    d = out["d"] * np.exp(-1j * np.angle(out["cd"]))

    # delta1 = a.real * IFO.a.node.qx.w
    # Positive 1j*HG10 addition to a beam results in a far-field shift in the -x direction
    # Or
    # Positive 1j*HG10 addition rotates the beam anticlockwise in the beams left-hadnded
    # coordinate system
    gamma1 = -a.imag * IFO.lambda0 / n1 / (np.pi * IFO.a.node.qx.w)
    # delta2 = b.real * IFO.b.node.qx.w
    gamma2 = -b.imag * IFO.lambda0 / n2 / (np.pi * IFO.b.node.qx.w)
    delta3 = c.real * IFO.c.node.qx.w
    # gamma3 = -c.imag * IFO.lambda0 / n2 / (np.pi * IFO.c.node.qx.w)
    # delta4 = d.real * IFO.d.node.qx.w
    gamma4 = -d.imag * IFO.lambda0 / n1 / (np.pi * IFO.d.node.qx.w)

    # correct refraction of angle on transmission through
    # surface
    assert abs(gamma2 - (gamma1 * n1 / n2)) < 1e-14
    # correct displacement from angle after propagation
    assert abs(delta3 - (gamma2 * 0.1)) < 1e-14
    # angle going in should be angle going out
    assert abs(gamma1 - gamma4) < 1e-14


def test_misaligned_input_beam_x_flip_coord():
    """For a misaligned incident beam the reflected field should be flipped LR, or have
    the odd modes with negative sign."""
    model = finesse.Model()
    model.parse(
        """
        l l1 P=1
        m m1 R=1 T=0
        l l2 P=1
        link(l1, 1000, m1, 1000, l2)
        fd Ei1 m1.p1.i 0
        fd Er1 m1.p1.o 0
        fd Ei2 m1.p2.i 0
        fd Er2 m1.p2.o 0
        modes(x, maxtem=1)
        gauss g1 m1.p1.i w0=1e-3 z=0
        """
    )
    model.beam_trace()
    model.m1.Rc = model.m1.p1.i.qx.Rc
    E = np.zeros(len(model.homs), dtype=complex)
    E[0] = 1
    K = make_bayerhelms_matrix(
        model.m1.p1.o.qx,
        model.m1.p1.o.qx,
        model.m1.p1.o.qy,
        model.m1.p1.o.qy,
        1e-6,
        0,
        select=model.homs,
    )
    E = K.data @ E
    model.l1.set_output_field(E, model.homs)
    model.l1.add_gouy_phase = False
    model.m1.xbeta = 0
    out = model.run()

    assert np.allclose(out["Ei1"][1], -out["Er1"][1], atol=1e-13, rtol=0)
    assert np.allclose(out["Ei2"][1], -out["Er2"][1], atol=1e-13, rtol=0)


def test_misalgined_input_beam_correction():
    """For a misaligned input beam the mirror is tilted such that the reflected beam is
    now aligned to the optical axis."""
    model = finesse.Model()
    model.parse(
        """
        l l1 P=1
        m m1 R=1 T=0
        l l2 P=1
        link(l1, 1000, m1, 1000, l2)
        fd Ei1 m1.p1.i 0
        fd Er1 m1.p1.o 0
        fd Ei2 m1.p2.i 0
        fd Er2 m1.p2.o 0
        modes(x, maxtem=2)
        gauss g1 m1.p1.i w0=1e-3 z=0
        """
    )
    model.beam_trace()
    model.m1.Rc = model.m1.p1.i.qx.Rc
    E = np.zeros(len(model.homs), dtype=complex)
    E[0] = 1
    K = make_bayerhelms_matrix(
        model.m1.p1.o.qx,
        model.m1.p1.o.qx,
        model.m1.p1.o.qy,
        model.m1.p1.o.qy,
        1e-6,
        0,
        select=model.homs,
    )
    E = K.data @ E
    model.l1.set_output_field(E, model.homs)
    model.l1.add_gouy_phase = True
    model.l2.set_output_field(E, model.homs)
    model.l2.add_gouy_phase = True
    model.m1.xbeta = -1e-6 / 2
    out = model.run()

    assert np.allclose(out["Er1"][0], 1, atol=1e-14, rtol=0)
    assert np.allclose(out["Er2"][0], 1, atol=1e-14, rtol=0)



from hypothesis import given, settings, note, HealthCheck
from testutils.text import stringify_list
from testutils.fuzzing import DEADLINE, recursive_arrays


@given(expected=recursive_arrays(operations=True))
@settings(
    deadline=DEADLINE,
    suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow],
)
def test_array_fuzzing(fuzz_argument, expected):
    array = stringify_list(expected)
    note(f"Array: {array}")
    assert fuzz_argument(array) == eval(array)



"""ABCD matrix result tests."""

import pytest
import numpy as np
from finesse.components import Beamsplitter, Lens, Mirror, Nothing, AstigmaticLens
from finesse.exceptions import TotalReflectionError, NoCouplingError


def abcd_space(L, nr):
    """ABCD matrix for a space."""
    return np.array([[1.0, L / nr], [0.0, 1.0]])


def abcd_lens(f):
    """ABCD matrix for a lens."""
    return np.array([[1.0, 0.0], [-1.0 / f, 1.0]])


def abcd_mirror_t(nr1, nr2, rc):
    """ABCD matrix for mirror transmission."""
    return np.array([[1.0, 0.0], [(nr2 - nr1) / rc, 1.0]])


def abcd_mirror_rt(nr1, rc):
    """ABCD matrix for mirror reflection."""
    # -1 for coordinate system flip
    return -1 * np.array([[1.0, 0.0], [-2 * nr1 / rc, 1.0]])


def abcd_mirror_rs(nr1, rc):
    """ABCD matrix for mirror reflection."""
    return np.array([[1.0, 0.0], [-2 * nr1 / rc, 1.0]])


def abcd_beamsplitter_tt(nr1, nr2, alpha1, alpha2, rc):
    """ABCD matrix for beam splitter tangential transmission."""
    dn = (nr2 * np.cos(alpha2) - nr1 * np.cos(alpha1)) / (
        np.cos(alpha1) * np.cos(alpha2)
    )
    return np.array(
        [
            [np.cos(alpha2) / np.cos(alpha1), 0.0],
            [dn / rc, np.cos(alpha1) / np.cos(alpha2)],
        ]
    )


def abcd_beamsplitter_ts(nr1, nr2, alpha1, alpha2, rc):
    """ABCD matrix for beam splitter transmission."""
    dn = (nr2 * np.cos(alpha2) - nr1 * np.cos(alpha1)) / (
        np.cos(alpha1) * np.cos(alpha2)
    )
    return np.array([[1.0, 0.0], [dn / rc, 1.0]])


def abcd_beamsplitter_rt(nr1, alpha1, rc):
    """ABCD matrix for beam splitter tangential reflection."""
    # -1 for coordinate system flip
    return -1 * np.array([[1.0, 0.0], [-2 * nr1 / (rc * np.cos(alpha1)), 1.0]])


def abcd_beamsplitter_rs(nr1, alpha1, rc):
    """ABCD matrix for beam splitter reflection."""
    return np.array([[1.0, 0.0], [-2 * nr1 * np.cos(alpha1) / rc, 1.0]])


def test_space_abcd_refractive_index_unity_transmit(model):
    """Test space ABCD matrix for refractive index = 1."""
    # length = 1 m, nr = 1.0
    L1 = 1.0
    nr = 1.0

    model.add([Nothing("NULL1"), Nothing("NULL2")])
    model.connect(model.NULL1.p2, model.NULL2.p1, name="S", L=L1, nr=nr)

    result = model.elements["S"].ABCD(model.NULL1.p2.o, model.NULL2.p1.i)
    target = abcd_space(L1, nr)
    assert np.allclose(result, target)

    # Check that reversing propagation direction yields same result.
    result_transmit_reversed = model.elements["S"].ABCD(
        model.NULL2.p1.o, model.NULL1.p2.i
    )
    assert np.allclose(result_transmit_reversed, target)

    # length = 34.5544 m, nr = 1.0
    L2 = 34.5544
    model.elements["S"].L = L2
    result_L2 = model.elements["S"].ABCD(model.NULL1.p2.o, model.NULL2.p1.i)
    target_L2 = abcd_space(L2, nr)
    assert np.allclose(result_L2, target_L2)


def test_space_abcd_refractive_index_non_unity_transmit(model):
    """Test space ABCD matrix for refractive index != 1."""
    # length = 2 m, nr = 1.44
    L = 2.0
    nr = 1.44

    model.add([Nothing("NULL1"), Nothing("NULL2")])
    model.connect(model.NULL1.p2, model.NULL2.p1, name="S", L=L, nr=nr)

    result = model.elements["S"].ABCD(model.NULL1.p2.o, model.NULL2.p1.i)
    target = abcd_space(L, nr)

    assert np.allclose(result, target)

    # Check that reversing propagation direction yields same result.
    result_transmit_reversed = model.elements["S"].ABCD(
        model.NULL2.p1.o, model.NULL1.p2.i
    )
    assert np.allclose(result_transmit_reversed, target)


def test_space_abcd_reflect(model):
    """Test space ABCD matrix reflection."""
    model.add([Nothing("NULL1"), Nothing("NULL2")])
    model.connect(model.NULL1.p2, model.NULL2.p1, name="S", L=1.0, nr=1.0)

    with pytest.raises(NoCouplingError):
        model.elements["S"].ABCD(model.NULL1.p2.o, model.NULL1.p2.i)


def test_lens_abcd_transmit():
    """Test lens ABCD matrix on transmission."""
    # focal length = 11 m
    f1 = 11.0
    lens = Lens("lens", f1)
    result = lens.ABCD(lens.p1.i, lens.p2.o)
    target = abcd_lens(f1)
    assert np.allclose(result, target)

    # Check that reversing propagation direction yields same result.
    result_transmit_reversed = lens.ABCD(lens.p2.i, lens.p1.o)
    assert np.allclose(result_transmit_reversed, target)

    # focal length = 34.5km
    f2 = 34.5e3
    lens.f = f2
    result_f2 = lens.ABCD(lens.p1.i, lens.p2.o)
    target_f2 = abcd_lens(f2)
    assert np.allclose(result_f2, target_f2)


def test_astigmatic_lens_abcd_transmit():
    """Test astigmatic lens ABCD matrix on transmission."""
    # focal lengths fx = 10 m, fy = 15 m
    fx = 10.0
    fy = 15.0
    lens = AstigmaticLens("astigmatic_lens", fx, fy)

    # Test for x direction
    result_x = lens.ABCD(lens.p1.i, lens.p2.o, direction="x")
    target_x = abcd_lens(fx)
    assert np.allclose(result_x, target_x)

    # Test for y direction
    result_y = lens.ABCD(lens.p1.i, lens.p2.o, direction="y")
    target_y = abcd_lens(fy)
    assert np.allclose(result_y, target_y)

    # Check that reversing propagation direction yields same result for x direction
    result_transmit_reversed_x = lens.ABCD(lens.p2.i, lens.p1.o, direction="x")
    assert np.allclose(result_transmit_reversed_x, target_x)

    # Check that reversing propagation direction yields same result for y direction
    result_transmit_reversed_y = lens.ABCD(lens.p2.i, lens.p1.o, direction="y")
    assert np.allclose(result_transmit_reversed_y, target_y)


def test_lens_abcd_reflect():
    """Test lens ABCD matrix on reflection."""
    lens = Lens("lens", 1.0)
    with pytest.raises(NoCouplingError):
        lens.ABCD(lens.p1.i, lens.p1.o)


def test_astigmatic_lens_abcd_reflect():
    """Test astigmatic lens ABCD matrix on reflection."""
    # focal lengths fx = 10 m, fy = 15 m
    fx = 10.0
    fy = 15.0
    lens = AstigmaticLens("astigmatic_lens", fx, fy)

    # Test for x direction
    with pytest.raises(NoCouplingError):
        lens.ABCD(lens.p1.i, lens.p1.o, direction="x")

    # Test for y direction
    with pytest.raises(NoCouplingError):
        lens.ABCD(lens.p1.i, lens.p1.o, direction="y")


def test_mirror_abcd_transmission_no_spaces():
    """Test mirror ABCD matrix on transmission with no attached spaces."""
    # radius of curvature = 2.5m, nr1 = nr2 = 1.0
    RoC = 2.5
    mirror = Mirror("M", Rc=RoC)
    result = mirror.ABCD(mirror.p1.i, mirror.p2.o)
    target = abcd_mirror_t(1.0, 1.0, RoC)
    assert np.allclose(result, target)

    # Check reversed propagation.
    result_transmit_reversed = mirror.ABCD(mirror.p2.i, mirror.p1.o)
    target_transmit_reversed = abcd_mirror_t(1.0, 1.0, -RoC)
    assert np.allclose(result_transmit_reversed, target_transmit_reversed)

    # radius of curvature = -12.6m, nr1 = nr2 = 1.0
    RoC = -12.6
    mirror.Rc = RoC
    result_neg_roc = mirror.ABCD(mirror.p1.i, mirror.p2.o)
    target_neg_roc = abcd_mirror_t(1.0, 1.0, RoC)
    assert np.allclose(result_neg_roc, target_neg_roc)


@pytest.mark.parametrize("xy,abcd", [("x", abcd_mirror_rt), ("y", abcd_mirror_rs)])
def test_mirror_abcd_reflection_no_spaces(xy, abcd):
    """Test mirror ABCD matrix on reflection with no attached spaces."""
    # radius of curvature = 7.7m, nr1 = nr2 = 1.0
    RoC = 7.7
    mirror = Mirror("M", Rc=RoC)
    result = mirror.ABCD(mirror.p1.i, mirror.p1.o, direction=xy)
    target = abcd(1.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_refl_reversed = mirror.ABCD(mirror.p2.i, mirror.p2.o, direction=xy)
    target_refl_reversed = abcd(1.0, -RoC)
    assert np.allclose(result_refl_reversed, target_refl_reversed)

    # radius of curvature = -5.5m, nr1 = nr2 = 1.0
    RoC = -5.5
    mirror.Rc = RoC
    result_neg_roc = mirror.ABCD(mirror.p1.i, mirror.p1.o, direction=xy)
    target_neg_roc = abcd(1.0, RoC)
    assert np.allclose(result_neg_roc, target_neg_roc)


def test_mirror_abcd_transmission_space_nr_unity_at_port1(model):
    """Test mirror ABCD matrix on transmission with attached space (nr = 1) at first
    port."""
    # radius of curvature = 3.4m, nr1 = nr2 = 1.0
    RoC = 3.4
    model.chain(Nothing("NULL"), 1, Mirror("M", Rc=RoC))
    result = model.M.ABCD(model.M.p1.i, model.M.p2.o)
    target = abcd_mirror_t(1.0, 1.0, RoC)
    assert np.allclose(result, target)


def test_mirror_abcd_transmission_space_nr_nonunity_at_port1(model):
    """Test mirror ABCD matrix on transmission with attached space (nr != 1) at first
    port."""
    # radius of curvature = -37.99m, nr1 = 1.08, nr2 = 1.0
    RoC = -37.99
    nr1 = 1.08
    nr2 = 1
    model.chain(Nothing("NULL"), {"L": 1, "nr": nr1}, Mirror("M", Rc=RoC))
    result = model.M.ABCD(model.M.p1.i, model.M.p2.o)
    target = abcd_mirror_t(nr1, nr2, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_trns_reversed = model.M.ABCD(model.M.p2.i, model.M.p1.o)
    target_trns_reversed = abcd_mirror_t(nr2, nr1, -RoC)
    assert np.allclose(result_trns_reversed, target_trns_reversed)


def test_mirror_abcd_reflection_space_nr_unity_at_port1(model):
    """Test mirror ABCD matrix on reflection with attached space (nr = 1) at first
    port."""
    # radius of curvature = 4.2m, nr1 = nr2 = 1.0
    RoC = 4.2
    model.chain(Nothing("NULL"), 1, Mirror("M", Rc=RoC))
    result = model.M.ABCD(model.M.p1.i, model.M.p1.o, direction="x")
    target = abcd_mirror_rt(1.0, RoC)
    assert np.allclose(result, target)


@pytest.mark.parametrize("xy,abcd", [("x", abcd_mirror_rt), ("y", abcd_mirror_rs)])
def test_mirror_abcd_reflection_space_nr_nonunity_at_port1(model, xy, abcd):
    """Test mirror ABCD matrix on reflection with attached space (nr != 1) at first
    port."""
    # radius of curvature = -14.7m, nr1 = 1.3, nr2 = 1.0
    RoC = -14.7
    nr1 = 1.3
    model.chain(Nothing("NULL"), {"L": 1, "nr": nr1}, Mirror("M", Rc=RoC))
    result = model.M.ABCD(model.M.p1.i, model.M.p1.o, direction=xy)
    target = abcd(nr1, RoC)
    assert np.allclose(result, target)

    # Check reversed propagation.
    result_refl_reversed = model.M.ABCD(model.M.p2.i, model.M.p2.o, direction=xy)
    target_refl_reversed = abcd(1.0, -RoC)
    assert np.allclose(result_refl_reversed, target_refl_reversed)


def test_mirror_abcd_transmission_spaces_nr_nonunity(model):
    """Test mirror ABCD matrix on transmission - attached spaces at both ports with (nr!=1)."""
    # radius of curvature = 17.67m, nr1 = 1.04, nr2 = 1.45
    RoC = 17.67
    nr1 = 1.04
    nr2 = 1.45
    model.chain(
        Nothing("NULL1"),
        {"L": 1, "nr": nr1},
        Mirror("M", Rc=RoC),
        {"L": 1, "nr": nr2},
        Nothing("NULL2"),
    )
    result = model.M.ABCD(model.M.p1.i, model.M.p2.o)
    target = abcd_mirror_t(nr1, nr2, RoC)
    assert np.allclose(result, target)

    # Check reversed propagation.
    result_trns_reversed = model.M.ABCD(model.M.p2.i, model.M.p1.o)
    target_trns_reversed = abcd_mirror_t(nr2, nr1, -RoC)
    assert np.allclose(result_trns_reversed, target_trns_reversed)


@pytest.mark.parametrize("xy,abcd", [("x", abcd_mirror_rt), ("y", abcd_mirror_rs)])
def test_mirror_abcd_reflection_spaces_nr_nonunity(model, xy, abcd):
    """Test mirror ABCD matrix on reflection - attached spaces at both ports with (nr!=1)."""
    # radius of curvature = -0.51m, nr1 = 1.11, nr2 = 1.55
    RoC = -0.51
    nr1 = 1.11
    nr2 = 1.55
    model.chain(
        Nothing("NULL1"),
        {"L": 1, "nr": nr1},
        Mirror("M", Rc=RoC),
        {"L": 1, "nr": nr2},
        Nothing("NULL2"),
    )
    result = model.M.ABCD(model.M.p1.i, model.M.p1.o, direction=xy)
    target = abcd(nr1, RoC)
    assert np.allclose(result, target)

    # Check reversed propagation.
    result_refl_reversed = model.M.ABCD(model.M.p2.i, model.M.p2.o, direction=xy)
    target_refl_reversed = abcd(nr2, -RoC)
    assert np.allclose(result_refl_reversed, target_refl_reversed)


def test_bs_abcd_no_coupling():
    """Test that beam splitter ABCD method raises an exception when calling with nodes
    that have no coupling."""
    bs = Beamsplitter("BS")

    with pytest.raises(NoCouplingError):
        bs.ABCD(bs.p1.i, bs.p4.o)
    with pytest.raises(NoCouplingError):
        bs.ABCD(bs.p2.i, bs.p3.o)
    with pytest.raises(NoCouplingError):
        bs.ABCD(bs.p3.i, bs.p2.o)
    with pytest.raises(NoCouplingError):
        bs.ABCD(bs.p4.i, bs.p1.o)


def test_flat_bs_abcd_transmission():
    """Test flat beam splitter ABCD matrix on transmission."""
    # radius of curvature = inf, nr1 = nr2 = 1.0
    bs = Beamsplitter("BS")
    result_t = bs.ABCD(bs.p1.i, bs.p3.o)
    target_t = abcd_beamsplitter_tt(1.0, 1.0, 0.0, 0.0, np.inf)
    assert np.allclose(result_t, target_t)

    result_s = bs.ABCD(bs.p1.i, bs.p3.o)
    target_s = abcd_beamsplitter_ts(1.0, 1.0, 0.0, 0.0, np.inf)
    assert np.allclose(result_s, target_s)


def test_flat_bs_abcd_reflection():
    """Test flat beam splitter ABCD matrix on transmission."""
    # radius of curvature = inf, nr1 = nr2 = 1.0
    bs = Beamsplitter("BS")
    result_t = bs.ABCD(bs.p1.i, bs.p2.o)
    target_t = abcd_beamsplitter_rt(1.0, 0.0, np.inf)
    assert np.allclose(result_t, target_t)

    result_s = bs.ABCD(bs.p1.i, bs.p3.o)
    target_s = abcd_beamsplitter_rs(1.0, 0.0, np.inf)
    assert np.allclose(result_s, target_s)


def test_bs_abcd_transmission_tangential_no_spaces():
    """Test beam splitter ABCD matrix on transmission for the tangential plane with no
    attached spaces."""
    # radius of curvature = 2.5m, nr1 = nr2 = 1.0
    RoC = 2.5
    bs = Beamsplitter("BS", Rc=RoC)
    result = bs.ABCD(bs.p1.i, bs.p3.o)
    target = abcd_beamsplitter_tt(1.0, 1.0, 0.0, 0.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_trns_reversed = bs.ABCD(bs.p3.i, bs.p1.o)
    target_trns_reversed = abcd_beamsplitter_tt(1.0, 1.0, 0.0, 0.0, -RoC)
    assert np.allclose(result_trns_reversed, target_trns_reversed)


def test_bs_abcd_transmission_sagittal_no_spaces():
    """Test beam splitter ABCD matrix on transmission for the sagittal plane with no
    attached spaces."""
    # radius of curvature = 13.8m, nr1 = nr2 = 1.0
    RoC = 13.8
    bs = Beamsplitter("BS", Rc=RoC)
    result = bs.ABCD(bs.p1.i, bs.p3.o, direction="y")
    target = abcd_beamsplitter_ts(1.0, 1.0, 0.0, 0.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_trns_reversed = bs.ABCD(bs.p3.i, bs.p1.o, direction="y")
    target_trns_reversed = abcd_beamsplitter_ts(1.0, 1.0, 0.0, 0.0, -RoC)
    assert np.allclose(result_trns_reversed, target_trns_reversed)


def test_bs_abcd_reflection_tangential_no_spaces():
    """Test beam splitter ABCD matrix on reflection for the tangential plane with no
    attached spaces."""
    # radius of curvature = -18.9m, nr1 = nr2 = 1.0
    RoC = -18.9
    bs = Beamsplitter("BS", Rc=RoC)
    result = bs.ABCD(bs.p1.i, bs.p2.o)
    target = abcd_beamsplitter_rt(1.0, 0.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_refl_reversed = bs.ABCD(bs.p2.i, bs.p1.o)
    target_refl_reversed = abcd_beamsplitter_rt(1.0, 0.0, RoC)
    assert np.allclose(result_refl_reversed, target_refl_reversed)


def test_bs_abcd_reflection_sagittal_no_spaces():
    """Test beam splitter ABCD matrix on reflection for the sagittal plane with no
    attached spaces."""
    # radius of curvature = 1.6m, nr1 = nr2 = 1.0
    RoC = 1.6
    bs = Beamsplitter("BS", Rc=RoC)
    result = bs.ABCD(bs.p1.i, bs.p2.o, direction="y")
    target = abcd_beamsplitter_rs(1.0, 0.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_refl_reversed = bs.ABCD(bs.p2.i, bs.p1.o, direction="y")
    target_refl_reversed = abcd_beamsplitter_rs(1.0, 0.0, RoC)
    assert np.allclose(result_refl_reversed, target_refl_reversed)


def test_bs_abcd_transmission_tangential_spaces(model):
    """Test beam splitter ABCD matrix on transmission for the tangential plane with
    attached spaces."""
    # radius of curvature = 19.94m, nr1 = 1.04, nr2 = 1.45
    RoC = 19.94
    nr1 = 1.04
    nr2 = 1.45
    model.chain(Nothing("NULL1"), {"L": 1, "nr": nr1}, Beamsplitter("BS", Rc=RoC))
    model.add(Nothing("NULL2"))
    model.connect(model.BS.p3, model.NULL2.p1, L=1, nr=nr2)
    result = model.BS.ABCD(model.BS.p1.i, model.BS.p3.o)
    target = abcd_beamsplitter_tt(nr1, nr2, 0.0, 0.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_trns_reversed = model.BS.ABCD(model.BS.p3.i, model.BS.p1.o)
    target_trns_reversed = abcd_beamsplitter_tt(nr2, nr1, 0.0, 0.0, -RoC)
    assert np.allclose(result_trns_reversed, target_trns_reversed)


def test_bs_abcd_transmission_sagittal_spaces(model):
    """Test beam splitter ABCD matrix on transmission for the sagittal plane with
    attached spaces."""
    # radius of curvature = -9m, nr1 = 1.33, nr2 = 1.044
    RoC = -9.0
    nr1 = 1.33
    nr2 = 1.044
    model.chain(Nothing("NULL1"), {"L": 1, "nr": nr1}, Beamsplitter("BS", Rc=RoC))
    model.add(Nothing("NULL2"))
    model.connect(model.BS.p3, model.NULL2.p1, L=1, nr=nr2)
    result = model.BS.ABCD(model.BS.p1.i, model.BS.p3.o, direction="y")
    target = abcd_beamsplitter_ts(nr1, nr2, 0.0, 0.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_trns_reversed = model.BS.ABCD(model.BS.p3.i, model.BS.p1.o, direction="y")
    target_trns_reversed = abcd_beamsplitter_tt(nr2, nr1, 0.0, 0.0, -RoC)
    assert np.allclose(result_trns_reversed, target_trns_reversed)


def test_bs_abcd_reflection_tangential_spaces(model):
    """Test beam splitter ABCD matrix on reflection for the tangential plane with
    attached spaces."""
    # radius of curvature = 10.66m, nr1 = 1.07, nr2 = 1.70
    RoC = 10.66
    nr1 = 1.70
    nr2 = 1.70
    model.chain(Nothing("NULL1"), {"L": 1, "nr": nr1}, Beamsplitter("BS", Rc=RoC))
    model.add(Nothing("NULL2"))
    model.connect(model.BS.p2, model.NULL2.p1, L=1, nr=nr2)
    result = model.BS.ABCD(model.BS.p1.i, model.BS.p2.o)
    target = abcd_beamsplitter_rt(nr1, 0.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_trns_reversed = model.BS.ABCD(model.BS.p2.i, model.BS.p1.o)
    target_trns_reversed = abcd_beamsplitter_rt(nr2, 0.0, RoC)
    assert np.allclose(result_trns_reversed, target_trns_reversed)


def test_bs_abcd_reflection_sagittal_spaces(model):
    """Test beam splitter ABCD matrix on reflection for the sagittal plane with attached
    spaces."""
    # radius of curvature = 10.66m, nr1 = 1.07, nr2 = 1.70
    RoC = 10.66
    nr1 = 1.07
    nr2 = 1.07
    model.chain(Nothing("NULL1"), {"L": 1, "nr": nr1}, Beamsplitter("BS", Rc=RoC))
    model.add(Nothing("NULL2"))
    model.connect(model.BS.p2, model.NULL2.p1, L=1, nr=nr2)
    result = model.BS.ABCD(model.BS.p1.i, model.BS.p2.o, direction="y")
    target = abcd_beamsplitter_rs(nr1, 0.0, RoC)
    assert np.allclose(result, target)

    # check reversed propagation
    result_trns_reversed = model.BS.ABCD(model.BS.p2.i, model.BS.p1.o, direction="y")
    target_trns_reversed = abcd_beamsplitter_rs(nr2, 0.0, RoC)
    assert np.allclose(result_trns_reversed, target_trns_reversed)


def test_bs_abcd_total_reflection(model):
    """Test beam splitter ABCD matrix with angles and refractive indices configured for
    total reflection."""
    nr1 = 1.45
    nr2 = 1.0
    alpha1 = 45.0
    model.chain(Nothing("NULL1"), {"L": 1, "nr": nr1}, Beamsplitter("BS", alpha=alpha1))
    model.add(Nothing("NULL2"))
    model.connect(model.BS.p3, model.NULL2.p1, L=1, nr=nr2)

    with pytest.raises(TotalReflectionError):
        model.BS.ABCD(model.BS.p1.i, model.BS.p3.o)



from finesse.knm.matrix cimport KnmMatrix
from finesse.knm.bayerhelms cimport (
    knm_bh_workspace,
    knm_bh_ws_init,
    knm_bh_ws_free,
    knm_bh_ws_is_changing,
    knm_bh_ws_recompute_mismatch,
    knm_bh_ws_recompute,
    knm_bh_ws_recompute_misalignment,
)



from finesse.cymath cimport complex_t
from finesse.tracing.tree cimport TraceTree
from finesse.tracing.forest cimport TraceForest
from finesse.simulations.base cimport ModelSettings, NodeBeamParam, SimConfigData
from finesse.frequency cimport frequency_info_t, FrequencyContainer, Frequency
from cpython.ref cimport PyObject
from .simulation cimport CNodeInfo


cdef class MatrixSystemWorkspaces:
    cdef readonly:
        list to_initial_fill
        list to_refill
        list to_rhs_refill
        list to_noise_refill
        list to_noise_input_refill
        list noise_detectors
        int num_to_refill
        int num_to_rhs_refill
        int num_to_noise_refill
        int num_to_noise_input_refill
        int num_noise_detectors
    cdef:
        PyObject** ptr_to_refill
        PyObject** ptr_to_rhs_refill
        PyObject** ptr_to_noise_refill
        PyObject** ptr_to_noise_input_refill
        PyObject** ptr_noise_detectors


cdef class BaseSolver:
    cdef:
        readonly MatrixSystemWorkspaces workspaces
        readonly dict connections
        readonly dict nodes
        readonly dict node_aliases
        readonly dict node_2_index
        readonly dict index_2_node

        Py_ssize_t num_nodes

        readonly FrequencyContainer optical_frequencies
        readonly dict signal_frequencies
        readonly tuple unique_elec_mech_fcnts # Unique frequency containers for mech/elec

        readonly dict noise_sources
        readonly bint any_frequencies_changing
        bint is_signal_matrix # TODO refactor to something not matrix named
        readonly bint forced_refill
        public bint manual_rhs # When true the RHS/input vector is
        readonly unsigned int num_solves # number of times solve has been called

        # Edges that have a changing mode mismatch, Node ID pair (in, out)
        readonly int[:, ::1] changing_mismatch_node_ids
        bint debug_mode

    cpdef setup_nodes(self, list nodes, dict node_aliases)
    cpdef clear_rhs(self)
    cpdef initial_fill(self)
    cpdef refill(self)
    cpdef refill_rhs(self)
    cpdef fill_rhs(self)
    cpdef fill_noise_inputs(self)
    cpdef refactor(self)
    cpdef factor(self)
    cpdef solve(self)
    cpdef solve_noises(self)
    cpdef construct(self)
    cpdef destruct(self)
    cpdef initial_run(self)
    cpdef run(self)

    cpdef tuple get_node_frequencies(self, node)

    cpdef update_frequency_info(self)
    cpdef assign_operators(self, workspaces)
    cpdef assign_noise_operators(self, workspaces)

    cpdef add_rhs(self, unicode key)

    cpdef Py_ssize_t findex(self, object node, Py_ssize_t freq)

    cpdef Py_ssize_t node_id(self, object node)
    cpdef get_node_info(self, name)



import logging
import weakref
import numpy as np


from finesse.cymath.cmatrix import _Column, _SubMatrix
from numpy.lib.stride_tricks import as_strided

LOGGER = logging.getLogger(__name__)


class DenseMatrix:
    """
    Examples
    --------
    Create a matrix with memory views of different submatrices for giving to components:

        > DM = DenseMatrix("abc")
        >
        > DM.declare_equations(5, 0, 'a')
        > DM.declare_equations(5, 1, 'b')
        > DM.declare_equations(5, 2, 'c')
        > DM.declare_equations(2, 3, 'd')
        >
        > v1 = DM.declare_submatrix_view(0, 1, 'b')
        > v2 = DM.declare_subdiagonal_view(0, 2, 'b')
        > v3 = DM.declare_submatrix_view(3, 1, 'b')
        >
        > DM.construct()
        >
        > v1[:] = 1
        > v2[:] = 0.5
        > v3[:] = 0.75
    """

    class SubMatrixView:
        """This class represents a sub-matrix view of a CCS sparse matrix. This allows
        code to access and set values without worrying about the underlying sparse
        compression being used. Although so far this is just for CCS formats.

        This object will get a view of a n-by-m sub-matrix starting at index (i,j). The
        values of his matrix will be set initially to the coordinates.
        """

        def __init__(self, Matrix, _from, _to, name, mtype):
            self.M = weakref.ref(Matrix)
            self._from = _from
            self._to = _to
            self.type = mtype
            Matrix._declare_submatrix(_from, _to, name, self, mtype)

        @property
        def from_idx(self):
            return self._from

        @property
        def to_idx(self):
            return self._to

        @property
        def view(self):
            return self.A

        def _updateview_(self):
            # This object represents a view of a matrix, this actually gets the
            # submatrix itself to use to fill things
            self.A = self.M().get_submatrix(self._from, self._to)

        def __getitem__(self, key):
            return self.A[key]

        def __setitem__(self, key, value):
            self.A[key] = value

    def __init__(self, name):
        self.__name = name
        self.sub_columns = {}
        self.num_eqs = 0
        self.diag_map = {}  # Maps submatrix index to RHS element index
        self.__callbacks = []

    @property
    def name(self):
        return self.__name

    def declare_submatrix_view(self, from_node, to_node, name):
        return DenseMatrix.SubMatrixView(self, from_node, to_node, name, "m")

    def declare_subdiagonal_view(self, from_node, to_node, name):
        return DenseMatrix.SubMatrixView(self, from_node, to_node, name, "d")

    def declare_equations(self, Neqs, index, name):
        """Adds a submatrix to the matrix along its diagonal. This defines what
        equations exist in the matrix, the submatrix values cannot be changed after
        this, they are always `1` Before other submatrices can be added to the matrix
        the diagonal must be specfied and how many equations it represents.

        Parameters
        ----------
        Neqs : Py_ssize_t
            Number of equations this submatrix represents
        _index : long
            Subcolumn index
        name : unicode
            Name used to indentify this coupling in the matrix for debugging
        """
        if index in self.sub_columns:
            raise Exception(
                "Diagonal elements already specified at index {}".format(index)
            )

        self.sub_columns[index] = _Column(Neqs, index, name)
        self.sub_columns[index].submatrices.append(
            _SubMatrix("d", Neqs, Neqs, index, name)
        )
        # Record what RHS index/number of equations this submatrix will start in
        self.diag_map[index] = self.num_eqs
        self.num_eqs += Neqs

    def add_block(self, Neqs, index, name):
        """
        Parameters
        ----------
        Neqs : Py_ssize_t
            Number of equations this submatrix represents
        _index : long
            Subcolumn index
        name : unicode
            Name used to indentify this coupling in the matrix for debugging
        """
        if index in self.sub_columns:
            raise Exception(
                "Diagonal elements already specified at index {}".format(index)
            )

        self.sub_columns[index] = _Column(Neqs, index, name)
        self.sub_columns[index].submatrices.append(
            _SubMatrix("m", Neqs, Neqs, index, name)
        )
        # Record what RHS index/number of equations this submatrix will start in
        self.diag_map[index] = self.num_eqs
        self.num_eqs += Neqs

    def _declare_submatrix(self, _from, _to, name, callback=None, type_="m"):
        """Adds a submatrix to the matrix. The nomenclature of `_from` and `_to` refer
        to the variable dependency of the equations this submatrix represents, i.e. the
        equations in submatrix `_to` depends on the values in `-from`. Therefore `_from`
        is the subcolumn index and `_to` is the subrow index.

        Parameters
        ----------
        _from : long
            Subcolumn index
        _to : long
            Subcolumn index
        name : unicode
            Name used to indentify this coupling in the matrix for debugging
        callback : function()
            A callback function that will be called once the matrix has been constructed
        type_ : char (optional)
            Either 'm' for a full submatrix or 'd' for a diagonal element only submatrix
        """
        if _from not in self.sub_columns:
            raise Exception(
                "Must add a diagonal submatrix at index {} first for this subcolumn".format(
                    _from
                )
            )
        if _to not in self.sub_columns:
            raise Exception(
                "Must add a diagonal submatrix at index {} first for this subcolumn".format(
                    _to
                )
            )

        _to_size = self.sub_columns[_to].size
        _from_size = self.sub_columns[_from].size

        self.sub_columns[_from].submatrices.append(
            _SubMatrix(type_, _from_size, _to_size, _to, name)
        )

        if callback:
            self.__callbacks.append(callback)

    def construct(self):
        """Constructing the matrix involves taking the metadata submatrix positions
        throughout the matrix and allocating the memory and building the various CCS
        matrix structures.

        After this the matrix can be populated and sovled.
        """
        self.M = -np.eye(self.num_eqs, dtype=complex)
        self.rhs = np.zeros((self.num_eqs, 1), dtype=complex)

        for cb in self.__callbacks:
            cb._updateview_()

    def get_submatrix(self, _from, _to):
        # _to -> row
        # _from -> column

        # Starting indicies
        sfidx = self.diag_map[_from]
        stidx = self.diag_map[_to]

        sm = [_ for _ in self.sub_columns[_from].submatrices if _.to == _to][0]
        to_size = sm.rows
        from_size = sm.columns

        # Only returns diagonal elements if needed
        if sm.type == "m":
            slf = slice(sfidx, sfidx + from_size, 1)
            slt = slice(stidx, stidx + to_size, 1)
            return self.M.view()[slt, slf]

        elif sm.type == "d":
            assert to_size == from_size
            return as_strided(
                self.M[stidx:, sfidx:],
                shape=(to_size,),
                strides=((self.num_eqs + 1) * self.M.itemsize,),
            )
        else:
            raise Exception(f"unexpected submatrix type {sm.type}")

    @property
    def num_equations(self):
        return self.num_eqs

    def get_matrix_elements(self):
        raise self.M

    def print_matrix(self):
        raise NotImplementedError()

    def print_rhs(self):
        raise NotImplementedError()

    def set_rhs(self, index, value):
        self.rhs[index] = value

    def clear_rhs(self):
        self.rhs[:] = 0

    def solve(self, transpose=False, conjugate=False):
        """Solve the matrix with options for transposing and conjugating.

        If `transpose` is False, solves the linear system :math:`Ax = b` using the
        ``Symbolic`` and ``Numeric`` objects stored by this class.

        Otherwise, solves the linear system :math:`A^T x = b` or :math:`A^H x = b`. The
        `conjugate` option is zero for :math:`A^T x = b` or non-zero for :math:`A^H x = b`.

        Parameters
        ----------
        transpose : bool
            Flag determining whether to solve the transpose of the matrix.

        conjugate : bool
            Flag determining whether to solve :math:`A^T x =b` or :math:`A^H x = b`
            for the transposed linear system.

        Returns
        -------
        out : np.ndarray
            The (negative) solution vector.
        """
        inv = np.linalg.inv(self.M)
        return inv @ self.rhs
